package ubiik.wp.controller;

import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicLong;
import java.util.stream.Collectors;
import ubiik.wp.model.devices.UUEndDevice;
import ubiik.wp.model.messages.PSMessage;
import ubiik.wp.model.messages.pscommands.PSCommand;
import ubiik.wp.model.messages.pscommands.PSCommandDelegate;
import ubiik.wp.model.messages.pscommands.PSCommandPingEndDevice;

public class InactiveDevicesManager implements Runnable, PSCommandDelegate {
  private static final int DEFAULT_PING_PAYLOAD_SIZE = 0;
  
  private static final long DEFAULT_CHECK_PERIOD_MILIS = 2000L;
  
  public static final long DEFAULT_PING_TIMEOUT = 60000L;
  
  public static final long DEFAULT_INACTIVE_TIME = 300000L;
  
  private AtomicLong checkPeriod;
  
  private AtomicLong inactivePeriod;
  
  private AtomicLong pingTimeout;
  
  private DevicesController devicesController;
  
  private MessageController messageController;
  
  private AtomicBoolean running;
  
  private Map<UUEndDevice, Long> pingsSent;
  
  public InactiveDevicesManager(long inactivePeriodMilis, long pingTimeoutMilis, DevicesController devicesController, MessageController messageController) {
    this.checkPeriod = new AtomicLong(2000L);
    this.inactivePeriod = new AtomicLong(inactivePeriodMilis);
    this.pingTimeout = new AtomicLong(pingTimeoutMilis);
    this.devicesController = devicesController;
    this.messageController = messageController;
    this.running = new AtomicBoolean(false);
    this.pingsSent = new HashMap<>();
  }
  
  public long getPingTimeout() {
    return this.pingTimeout.get();
  }
  
  public void setPingTimeout(long pingTimeout) {
    this.pingTimeout.set(pingTimeout);
  }
  
  public long getCheckPeriod() {
    return this.checkPeriod.get();
  }
  
  public void setCheckPeriod(long checkPeriod) {
    this.checkPeriod.set(checkPeriod);
  }
  
  public long getInactivePeriod() {
    return this.inactivePeriod.get();
  }
  
  public void setInactivePeriod(long period) {
    this.inactivePeriod.set(period);
  }
  
  public void stop() {
    this.running.set(false);
  }
  
  public void run() {
    this.pingsSent.clear();
    this.running.set(true);
    long currentDateTime = -1L;
    long inactiveT = 0L;
    while (this.running.get()) {
      currentDateTime = (new Date()).getTime();
      inactiveT = this.inactivePeriod.get();
      Collection<UUEndDevice> allDevices = this.devicesController.getDevices();
      List<UUEndDevice> devices = (List<UUEndDevice>)allDevices.stream().filter(ed -> (ed.getStatus() == 2 || ed.getStatus() == 1 || ed.getStatus() == 6)).collect(Collectors.toList());
      for (UUEndDevice device : devices) {
        if (currentDateTime > device.getLastActivity().getTime() + inactiveT) {
          if (!this.pingsSent.containsKey(device)) {
            PSCommandPingEndDevice cmd = new PSCommandPingEndDevice(this, device.getUuid(), 0, (int)this.devicesController.getNextPingId());
            this.messageController.sendMessage((PSMessage)cmd);
            this.devicesController.addPing(cmd);
            this.pingsSent.put(device, Long.valueOf(currentDateTime));
          } 
          continue;
        } 
        this.pingsSent.remove(device);
      } 
      Map<UUEndDevice, Long> updatedPingsSent = new HashMap<>();
      for (Map.Entry<UUEndDevice, Long> entry : this.pingsSent.entrySet()) {
        if (currentDateTime - ((Long)entry.getValue()).longValue() < this.pingTimeout.get()) {
          updatedPingsSent.put(entry.getKey(), entry.getValue());
          continue;
        } 
        UUEndDevice ed = entry.getKey();
        if (ed.getStatus() != 3)
          ed.setStatus((byte)-2); 
      } 
      this.pingsSent = updatedPingsSent;
      try {
        Thread.sleep(this.checkPeriod.get());
      } catch (InterruptedException e) {
        e.printStackTrace();
        this.running.set(false);
      } 
    } 
  }
  
  public synchronized boolean isRunning() {
    return this.running.get();
  }
  
  public void commandandReceivedResponse(PSCommand command) {}
}
