package ubiik.wp.controller;

import com.opencsv.CSVWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Base64;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import ubiik.wp.model.devices.UUEndDevice;
import ubiik.wp.model.messages.DownlinkAcknowledgement;
import ubiik.wp.model.messages.JSONMessageDelegate;
import ubiik.wp.model.messages.PSAckMessage;
import ubiik.wp.model.messages.PSDownlink;
import ubiik.wp.model.messages.PSLogMessage;
import ubiik.wp.model.messages.PSMessage;
import ubiik.wp.model.messages.PSMessageDeviceStatus;
import ubiik.wp.model.messages.PSMessageDiscovery;
import ubiik.wp.model.messages.PSMessageNewDiscovery;
import ubiik.wp.model.messages.PSMessageJson;
import ubiik.wp.model.messages.PSMessageResponse;
import ubiik.wp.model.messages.PSMulticastNotification;
import ubiik.wp.model.messages.PSNakMessage;
import ubiik.wp.model.messages.PSPong;
import ubiik.wp.model.messages.PSUplink;
import ubiik.wp.model.messages.UplinksReceiver;
import ubiik.wp.model.messages.pscommands.PSCommand;
import ubiik.wp.model.messages.pscommands.responses.PSCommandResponse;
import ubiik.wp.util.ApplicationLogger;

public class PSMessageStorage {
  private static final int MAX_STORED_MESSAGES = 1000;
  
  private static final int MAX_STORED_UPLINKS = 1000;
  
  private static final int MAX_STORED_DOWNLINKS = 1000;
  
  private static final int MAX_STORED_COMMANDS = 1000;
  
  private static final int MAX_STORED_LOG_MESSAGES = 1000;
  
  private List<MessagesObserver> observers;
  
  private List<UplinksReceiver> uplinksReceiver;
  
  private List<PSUplink> uplinks;
  
  private List<PSMessage> sentMessages;
  
  private long commandsCount;
  
  private LinkedHashMap<Long, PSDownlink> downlinks;
  
  private LinkedHashMap<Long, PSCommand> commands;
  
  private List<PSLogMessage> logMessages;
  
  private Map<String, JSONMessageDelegate> jsonMessageDelegates;
  
  private Object jsonMessageDelegatesLock = new Object();
  
  private DevicesController devicesController;
  
  private PSMessageStorage() {
    this.uplinks = new LinkedList<>();
    this.downlinks = new LinkedHashMap<>();
    this.commands = new LinkedHashMap<>();
    this.commandsCount = 0L;
    this.observers = new LinkedList<>();
    this.logMessages = new LinkedList<>();
    this.uplinksReceiver = new LinkedList<>();
    this.sentMessages = new LinkedList<>();
    this.jsonMessageDelegates = new HashMap<>();
    this.devicesController = null;
  }
  
  private static class Holder {
    static final PSMessageStorage INSTANCE = new PSMessageStorage();
  }
  
  public static PSMessageStorage getInstance() {
    return Holder.INSTANCE;
  }
  
  public synchronized DevicesController getDevicesController() {
    return this.devicesController;
  }
  
  public synchronized void setDevicesController(DevicesController devicesController) {
    this.devicesController = devicesController;
  }
  
  public void restart() {
    this.uplinks.clear();
    this.downlinks.clear();
    this.commands.clear();
    this.commandsCount = 0L;
    notifyObservers();
  }
  
  public void clearUplinks() {
    synchronized (this.uplinks) {
      this.uplinks.clear();
    } 
  }
  
  public void clearDownlinks() {
    synchronized (this.downlinks) {
      this.downlinks.clear();
    } 
  }
  
  public void addMessage(PSMessage message) {
    if (message instanceof PSUplink) {
      synchronized (this.uplinks) {
        PSUplink uplink = (PSUplink)message;
        this.uplinks.add(uplink);
        if (this.devicesController != null)
          this.devicesController.addUplinkToDevice(uplink); 
        if (this.uplinks.size() > 1000)
          this.uplinks.remove(0); 
        synchronized (this.uplinksReceiver) {
          for (UplinksReceiver receiver : this.uplinksReceiver)
            receiver.receiveUplink(uplink); 
        } 
      } 
    } else if (message instanceof PSDownlink) {
      synchronized (this.downlinks) {
        PSDownlink downlink = (PSDownlink)message;
        this.downlinks.put(Long.valueOf(downlink.getApplicationId()), downlink);
        if (this.devicesController != null)
          this.devicesController.addDownlinkToDevice(downlink); 
        if (this.downlinks.size() > 1000) {
          Map.Entry<Long, PSDownlink> entry = this.downlinks.entrySet().iterator().next();
          if (entry != null)
            this.downlinks.remove(entry.getKey()); 
        } 
      } 
    } else if (message instanceof DownlinkAcknowledgement) {
      addDownlinkAcknowledgement((DownlinkAcknowledgement)message);
    } else if (message instanceof PSCommandResponse) {
      PSCommandResponse response = (PSCommandResponse)message;
      addCommandResponse(response);
    } else if (message instanceof PSCommand) {
      synchronized (this.commands) {
        PSCommand command = (PSCommand)message;
        this.commands.put(Long.valueOf(command.getApplicationId()), command);
        this.commandsCount++;
        if (this.commands.size() > 1000) {
          Map.Entry<Long, PSCommand> entry = this.commands.entrySet().iterator().next();
          if (entry != null)
            this.commands.remove(entry.getKey()); 
        } 
      } 
    } else if (message instanceof PSMessageDiscovery) {
      if (this.devicesController != null)
        this.devicesController.setBsid(((PSMessageDiscovery)message).getDeviceId()); 
    } else if (message instanceof PSMessageNewDiscovery) {
      if (this.devicesController != null)
      {
        this.devicesController.setBsid(((PSMessageNewDiscovery)message).getDeviceUUID());
      }
    } else if (message instanceof PSMessageDeviceStatus) {
      PSMessageDeviceStatus deviceStatusMessage = (PSMessageDeviceStatus)message;
      if (deviceStatusMessage.isForEndDevice()) {
        if (this.devicesController != null)
        	if(deviceStatusMessage.getDeviceType() == deviceStatusMessage.DEVICE_TYPE_ED) {
        		//this.devicesController.
        		this.devicesController.setEndDeviceStatus(deviceStatusMessage.getDeviceId(), deviceStatusMessage.getStatus(), deviceStatusMessage
              .getJoinTime(), deviceStatusMessage.getRssi(), deviceStatusMessage.getLastActivityTime(), deviceStatusMessage
              .getConnectionTime(), deviceStatusMessage.getTotalULUNACK(), deviceStatusMessage.getTotalULACK(), deviceStatusMessage.getDeviceType(), deviceStatusMessage.getFirmwareVersion(), deviceStatusMessage.getJoinFoi());
        	} else if(deviceStatusMessage.getDeviceType() == deviceStatusMessage.DEVICE_TYPE_FOREIGN_ED) {
        		this.devicesController.setEndDeviceStatus(deviceStatusMessage.getDeviceId(), deviceStatusMessage.getStatus(), new Date(), 
        				(short)0, new Date(), deviceStatusMessage.getConnectionTime(), null, null, deviceStatusMessage.getDeviceType(), deviceStatusMessage.getFirmwareVersion());
        	} 
        	
      } else if (this.devicesController != null) {
        this.devicesController.setBsVersionAndBuildDate(deviceStatusMessage.getFirmwareVersion(), deviceStatusMessage.getBsBuildDateTime());
        this.devicesController.setNumEndDevices(deviceStatusMessage.getNumEndDevices());
      } 
    } else if (message instanceof PSLogMessage) {
      synchronized (this.logMessages) {
        this.logMessages.add((PSLogMessage)message);
        if (this.logMessages.size() > 1000)
          this.logMessages.remove(0); 
      } 
    } else if (message instanceof PSMessageJson) {
      addJsonMessage((PSMessageJson)message);
    } else if (message instanceof PSNakMessage) {
      addNak((PSNakMessage)message);
    } else if (message instanceof PSAckMessage) {
      addAck((PSAckMessage)message);
    } else if (message instanceof PSMulticastNotification) {
      handleMulticastNotification((PSMulticastNotification)message);
    } else if (message instanceof PSPong) {
      handlePSPong((PSPong)message);
    } 
    notifyObservers();
  }
  
  public void storeSentMessage(PSMessage message) {
    synchronized (this.sentMessages) {
      this.sentMessages.add(message);
      if (this.sentMessages.size() > 1000)
        this.sentMessages.remove(0); 
    } 
  }
  
  private void addCommandResponse(PSCommandResponse response) {
    synchronized (this.commands) {
      PSCommand cmd = this.commands.get(Long.valueOf(response.getApplicationId()));
      if (cmd != null)
        cmd.setResponse(response); 
    } 
  }
  
  private void addDownlinkAcknowledgement(DownlinkAcknowledgement dlAck) {
    synchronized (this.downlinks) {
      PSDownlink dl = this.downlinks.get(Long.valueOf(dlAck.getApplicationId()));
      if (dl != null)
        dl.setDowlinkAcknowledgement(dlAck); 
    } 
  }
  
  private void addNak(PSNakMessage nak) {
    addMessageResponse((PSMessageResponse)nak);
  }
  
  private void addAck(PSAckMessage ack) {
    addMessageResponse((PSMessageResponse)ack);
  }
  
  private void addMessageResponse(PSMessageResponse response) {
    this.sentMessages.stream()
      .filter(m -> (m.getMessageId() == response.getMessageIDtoACK())).findFirst()
      .ifPresent(val -> val.setResponse(response));
  }
  
  private void notifyObservers() {
    synchronized (this.observers) {
      for (MessagesObserver observer : this.observers)
        observer.update(); 
    } 
  }
  
  public void addObdserver(MessagesObserver ovserver) {
    synchronized (this.observers) {
      this.observers.add(ovserver);
    } 
  }
  
  public void removeObdserver(MessagesObserver ovserver) {
    synchronized (this.observers) {
      this.observers.remove(ovserver);
    } 
  }
  
  public void addUplinksReceiver(UplinksReceiver receiver) {
    synchronized (this.uplinksReceiver) {
      this.uplinksReceiver.add(receiver);
    } 
  }
  
  public void removeUplinksReceiver(UplinksReceiver receiver) {
    synchronized (this.uplinksReceiver) {
      this.uplinksReceiver.remove(receiver);
    } 
  }
  
  public List<PSUplink> getUplinks() {
    synchronized (this.uplinks) {
      return new LinkedList<>(this.uplinks);
    } 
  }
  
  public List<PSUplink> getUplinks(int max) {
    synchronized (this.uplinks) {
      int toIndex = this.uplinks.size();
      int fromIndex = (toIndex - max >= 0) ? (toIndex - max) : 0;
      return new LinkedList<>(this.uplinks.subList(fromIndex, toIndex));
    } 
  }
  
  public List<PSDownlink> getDownlinks() {
    synchronized (this.downlinks) {
      return new LinkedList<>(this.downlinks.values());
    } 
  }
  
  public List<PSDownlink> getDownlinks(int max) {
    synchronized (this.downlinks) {
      int toIndex = this.downlinks.size();
      int fromIndex = (toIndex - max >= 0) ? (toIndex - max) : 0;
      return (new LinkedList<>(this.downlinks.values())).subList(fromIndex, toIndex);
    } 
  }
  
  public List<PSCommand> getCommands(int max) {
    synchronized (this.commands) {
      int toIndex = this.commands.size();
      int fromIndex = (toIndex - max >= 0) ? (toIndex - max) : 0;
      return (new LinkedList<>(this.commands.values())).subList(fromIndex, toIndex);
    } 
  }
  
  public List<PSCommand> getCommands() {
    synchronized (this.commands) {
      return new LinkedList<>(this.commands.values());
    } 
  }
  
  public List<PSLogMessage> getLogs() {
    synchronized (this.logMessages) {
      return new LinkedList<>(this.logMessages);
    } 
  }
  
  public long getUplinksCount() {
    long uplinksCount = 0L;
    if (this.devicesController != null) {
      Collection<UUEndDevice> collection = this.devicesController.getDevices();
      for (UUEndDevice uuEndDevice : collection)
        uplinksCount += uuEndDevice.getUplinksCounter(); 
    } 
    return uplinksCount;
  }
  
  public long getDownlinksCount() {
    long downlinksCount = 0L;
    if (this.devicesController != null) {
      Collection<UUEndDevice> collection = this.devicesController.getDevices();
      for (UUEndDevice uuEndDevice : collection)
        downlinksCount += uuEndDevice.getDownlinksCounter(); 
    } 
    return downlinksCount;
  }
  
  public long getCommandsCount() {
    return this.commandsCount;
  }
  
  public void addJSONMessageDelegate(String key, JSONMessageDelegate delegate) {
    synchronized (this.jsonMessageDelegatesLock) {
      this.jsonMessageDelegates.put(key, delegate);
    } 
  }
  
  public void removeJSONMessageDelegate(String key) {
    synchronized (this.jsonMessageDelegatesLock) {
      this.jsonMessageDelegates.remove(key);
    } 
  }
  
  private void addJsonMessage(PSMessageJson message) {
    JSONObject root = message.getJsonObject();
    if (root != null)
      synchronized (this.jsonMessageDelegatesLock) {
        if (root.containsKey("database")) {
          JSONObject dbJson = (JSONObject)root.get("database");
          handleSQLQueryMessage(dbJson);
        } 
        if (root.containsKey("file")) {
          JSONObject jsonObj = (JSONObject)root.get("file");
          handleFileMessage(jsonObj);
        } 
        for (Map.Entry<String, JSONMessageDelegate> entry : this.jsonMessageDelegates.entrySet()) {
          if (root.containsKey(entry.getKey()))
            ((JSONMessageDelegate)entry.getValue()).parseJSON((JSONObject)root.get(entry.getKey())); 
        } 
      }  
  }
  
  private String getFileNameWithDate(String name) {
    Date startTime = Calendar.getInstance().getTime();
    DateFormat dateFormat = new SimpleDateFormat("yyyyMMdd_HHmmss");
    String fileName = dateFormat.format(startTime);
    fileName = fileName + "_" + fileName;
    return fileName;
  }
  
  private void handleFileMessage(JSONObject jsonObj) {
    if (jsonObj.containsKey("error"))
      return; 
    try {
      String filename;
      int offset = 0;
      if (jsonObj.containsKey("fragment_offset"))
        offset = ((Long)jsonObj.get("fragment_offset")).intValue(); 
      if (jsonObj.containsKey("filename")) {
        String name = (String)jsonObj.get("filename");
        int lastPosDirUnix = name.lastIndexOf("/");
        int lastPosDirWin = name.lastIndexOf("\\");
        int pos = Math.max(lastPosDirUnix, lastPosDirWin);
        pos = (pos >= 0) ? (pos + 1) : 0;
        name = name.substring(pos);
        filename = name;
      } else {
        filename = getFileNameWithDate("");
      } 
      if (jsonObj.containsKey("content")) {
        String contentb64 = (String)jsonObj.get("content");
        byte[] byteArray = Base64.getUrlDecoder().decode(contentb64);
        RandomAccessFile writer = new RandomAccessFile(filename, "rw");
        try {
          writer.seek(offset);
          writer.write(byteArray);
        } catch (Exception e) {
          e.printStackTrace();
        } finally {
          writer.close();
        } 
      } 
    } catch (Exception e) {
      ApplicationLogger.debug("Error receiving file from JSON object");
      ApplicationLogger.debug(e.toString());
      e.printStackTrace();
    } 
  }
  
  private void handleSQLQueryMessage(JSONObject jsonObj) {
    if (jsonObj.containsKey("error"))
      return; 
    CSVWriter csvWriter = null;
    String filename;
    try {
      if (jsonObj.containsKey("name")) {
        String name = (String)jsonObj.get("name");
        name = name.replaceAll("[^a-zA-Z0-9\\.\\-]", "_");
        if (name.length() > 20)
          name = name.substring(0, 20); 
        filename = getFileNameWithDate(name);
      } else {
        filename = getFileNameWithDate("");
      } 
      String csvfilename = filename + ".csv";
      File file = new File(csvfilename);
      FileWriter fWriter = new FileWriter(file);
      csvWriter = new CSVWriter(fWriter);
      if (jsonObj.containsKey("columns")) {
        JSONArray columnsJson = (JSONArray)jsonObj.get("columns");
        String[] header = new String[columnsJson.size()];
        for (int i = 0; i < columnsJson.size(); i++)
          header[i] = columnsJson.get(i).toString(); 
        csvWriter.writeNext(header);
      } 
      if (jsonObj.containsKey("results")) {
        JSONArray resultsJson = (JSONArray)jsonObj.get("results");
        String[] line = null;
        for (int i = 0; i < resultsJson.size(); i++) {
          line = resultsJson.get(i).toString().split(",");
          csvWriter.writeNext(line);
        } 
      } 
    } catch (Exception e) {
      ApplicationLogger.debug("Error receiving SQL response from JSON object");
      ApplicationLogger.debug(e.toString());
      e.printStackTrace();
    } finally {
      try {
        if (csvWriter != null) {
          csvWriter.flush();
          csvWriter.close();
        } 
      } catch (IOException e) {
        e.printStackTrace();
      } 
    } 
  }
  
  private void handlePSPong(PSPong message) {
    if (this.devicesController != null) {
      this.devicesController.addPong(message);
    } else {
      System.out.println("this.devicesController == null");
    } 
  }
  
  private void handleMulticastNotification(PSMulticastNotification message) {
    if (this.devicesController != null)
      if (message.getNotificationType() == 0) {
        this.devicesController.setMulticastGroups(message.getMulticastGroups(), message.getDevicesStatusInMulticastGroups());
      } else if (message.getNotificationType() == 1) {
        this.devicesController.addMulticastGroups(message.getMulticastGroups(), message.getDevicesStatusInMulticastGroups());
      }  
  }
}
