package ubiik.wp.demo.controller;

import com.opencsv.CSVWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.Optional;
import javafx.application.Platform;
import javafx.collections.FXCollections;
import javafx.collections.ListChangeListener;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.scene.chart.LineChart;
import javafx.scene.chart.XYChart;
import javafx.scene.control.TableColumn;
import javafx.scene.control.TableView;
import javafx.scene.control.cell.PropertyValueFactory;
import org.controlsfx.control.table.TableFilter;
import ubiik.wp.controller.DevicesController;
import ubiik.wp.controller.MessageController;
import ubiik.wp.demo.model.PhySample;
import ubiik.wp.model.messages.PSUplink;
import ubiik.wp.util.ApplicationLogger;

public class PhyStatisticsViewController extends UplinksDemoViewController {
  private ObservableList<PhySample> latestPhySamples = FXCollections.observableArrayList();
  
  private LinkedHashMap<String, ObservableList<XYChart.Data<Date, Number>>> rssiSeriesDataMap;
  
  private ObservableList<XYChart.Series<Date, Number>> rssiSeries;
  
  private ObservableList<XYChart.Series<Date, Number>> uplinkRssiSeries;
  
  private ObservableList<XYChart.Series<Date, Number>> downlinkRssiSeries;
  
  private LineChart<Date, Number> rssiChart;
  
  private LineChart<Date, Number> uplinkRssiChart;
  
  private LineChart<Date, Number> downlinkRssiChart;
  
  @FXML
  protected TableView<PhySample> tableView;
  
  private TableFilter<PhySample> filter;
  
  private CSVWriter csvWriter;
  
  private boolean useOneChartForAllSeries;
  
  public PhyStatisticsViewController(DevicesController devicesController, MessageController messageController) {
    super(devicesController, messageController);
    this.rssiSeriesDataMap = new LinkedHashMap<>();
    this.rssiSeries = FXCollections.observableArrayList();
    this.uplinkRssiSeries = FXCollections.observableArrayList();
    this.downlinkRssiSeries = FXCollections.observableArrayList();
    this.csvWriter = null;
    this.useOneChartForAllSeries = true;
  }
  
  private boolean setUpCsvFile() {
    Date startTime = Calendar.getInstance().getTime();
    boolean success = true;
    try {
      DateFormat dateFormat = new SimpleDateFormat("yyyyMMdd_HHmmss");
      String fileName = dateFormat.format(startTime) + ".csv";
      File file = new File(fileName);
      FileWriter fWriter = new FileWriter(file);
      this.csvWriter = new CSVWriter(fWriter);
      String[] header = { "UUEID", "Uplink RSSI", "Downlink RSSI", "Received Time", "Sample time", "Battery Voltage", "Temperature", "# Packets" };
      this.csvWriter.writeNext(header);
    } catch (IOException e) {
      success = false;
      this.csvWriter = null;
      e.printStackTrace();
    } 
    return success;
  }
  
  public void authenticationFinished(boolean autheticated) {
    super.authenticationFinished(autheticated);
    if (autheticated) {
      setUpCsvFile();
    } else {
      this.csvWriter = null;
    } 
  }
  
  public void disconnect() {
    super.disconnect();
    if (this.csvWriter != null)
      try {
        this.csvWriter.flush();
        this.csvWriter.close();
      } catch (IOException e) {
        e.printStackTrace();
      }  
  }
  
  private void writeSampleToCsv(PhySample sample) {
    if (this.csvWriter != null) {
      String[] stringList = { sample.getUuid(), sample.getUplinkRssiStr(), Integer.toString(sample.getRssi()), sample.getUplinkTimeStr(), sample.getSampleTimeStr(), Integer.toString(sample.getBatteryVoltage()), Integer.toString(sample.getTemperature()), Integer.toString(sample.getPackets()) };
      this.csvWriter.writeNext(stringList);
    } 
  }
  
  public void initialize() {
    super.initialize();
    setUpRssiChart();
    setUpSamplesTable();
    setTitle("Phy Statistics");
    this.filter = TableFilter.forTableView(this.tableView).apply();
    this.filter.getFilteredList().addListener(new ListChangeListener<PhySample>() {
          public void onChanged(ListChangeListener.Change<? extends PhySample> c) {
            PhyStatisticsViewController.this.updateSeries();
          }
        });
  }
  
  private void updateSeries() {
    this.rssiSeries.clear();
    this.uplinkRssiSeries.clear();
    this.downlinkRssiSeries.clear();
    for (PhySample sample : this.filter.getFilteredList()) {
      String uuid = sample.getUuid();
      ObservableList<XYChart.Data<Date, Number>> downlinksSeries = this.rssiSeriesDataMap.get("D-" + uuid);
      ObservableList<XYChart.Data<Date, Number>> uplinksSeries = this.rssiSeriesDataMap.get("U-" + uuid);
      if (uplinksSeries != null) {
        XYChart.Series<Date, Number> ulS = new XYChart.Series<>("U-" + uuid, uplinksSeries);
        this.rssiSeries.add(ulS);
        this.uplinkRssiSeries.add(ulS);
      } 
      if (downlinksSeries != null) {
        XYChart.Series<Date, Number> dlS = new XYChart.Series<>("D-" + uuid, downlinksSeries);
        this.rssiSeries.add(dlS);
        this.downlinkRssiSeries.add(dlS);
      } 
    } 
  }
  
  private void setUpRssiChart() {
    if (this.useOneChartForAllSeries) {
      this.rssiChart = setUpLineChart(this.rssiChart, this.rssiSeries, "Uplink/Downlink RSSI");
    } else {
      this.uplinkRssiChart = setUpLineChart(this.uplinkRssiChart, this.uplinkRssiSeries, "Uplink RSSI");
      this.downlinkRssiChart = setUpLineChart(this.downlinkRssiChart, this.downlinkRssiSeries, "Downlink RSSI");
    } 
  }
  
  private void setUpSamplesTable() {
    this.tableView.getColumns().clear();
    TableColumn<PhySample, String> deviceCol = new TableColumn<>("End Device");
    TableColumn<PhySample, Integer> appRssiCol = new TableColumn<>("Downlink RSSI");
    TableColumn<PhySample, Integer> uplinkRssiCol = new TableColumn<>("Uplink RSSI");
    TableColumn<PhySample, String> datelCol = new TableColumn<>("Date");
    deviceCol.prefWidthProperty().set(350.0D);
    appRssiCol.prefWidthProperty().set(200.0D);
    uplinkRssiCol.prefWidthProperty().set(200.0D);
    datelCol.prefWidthProperty().set(250.0D);
    deviceCol.setSortable(true);
    appRssiCol.setSortable(false);
    uplinkRssiCol.setSortable(false);
    datelCol.setSortable(false);
    deviceCol.setCellValueFactory(new PropertyValueFactory<>("uuid"));
    appRssiCol.setCellValueFactory(new PropertyValueFactory<>("rssi"));
    uplinkRssiCol.setCellValueFactory(new PropertyValueFactory<>("uplinkRssi"));
    datelCol.setCellValueFactory(new PropertyValueFactory<>("uplinkTimeStr"));
    this.tableView.getColumns().addAll(new TableColumn[] { deviceCol, appRssiCol, uplinkRssiCol, datelCol });
    this.tableView.setItems(this.latestPhySamples);
  }
  
  private PhySample getPhyReadingFromUplink(PSUplink uplink) {
    PhySample phySample = null;
    byte[] data = uplink.getApplicationData();
    try {
      if (data.length == 13) {
        ByteBuffer buffer = ByteBuffer.wrap(data).order(ByteOrder.LITTLE_ENDIAN);
        byte appType = buffer.get();
        if (appType == 32) {
          int npkt = 0xFFFF & buffer.getShort();
          int rssi = buffer.getShort();
          int batteryVoltage = buffer.getShort();
          int temperature = buffer.getShort();
          long intSampleTime = buffer.getInt();
          Date sampleTime = new Date(intSampleTime * 1000L);
          ApplicationLogger.debug("Sample time: " + intSampleTime + " " + sampleTime.toString());
          phySample = new PhySample(uplink.getTargetString(), uplink.getTimestamp(), uplink.getRssi(), npkt, rssi, batteryVoltage, temperature, uplink.getCreationDate(), sampleTime);
          ApplicationLogger.debug("New sample: " + phySample.toString());
        } 
      } 
    } catch (Exception e) {
      e.printStackTrace();
    } 
    return phySample;
  }
  
  protected void processUplink(PSUplink uplink) {
    final PhySample sample = getPhyReadingFromUplink(uplink);
    if (sample != null) {
      final String target = uplink.getTargetString();
      Platform.runLater(new Runnable() {
            public void run() {
              boolean newSeries = PhyStatisticsViewController.this.addUplinkRssiSample(sample, target);
              PhyStatisticsViewController.this.updateLatestSamples(sample);
              if (newSeries)
                PhyStatisticsViewController.this.updateSeries(); 
            }
          });
      writeSampleToCsv(sample);
    } 
  }
  
  private boolean addUplinkRssiSample(PhySample sample, String uuid) {
    boolean newSeries = false;
    if (sample.getUplinkRssi() != null && 
      this.rssiSeriesDataMap.get("U-" + uuid) == null) {
      ObservableList<XYChart.Data<Date, Number>> seriesDataU = FXCollections.observableArrayList();
      this.rssiSeriesDataMap.put("U-" + uuid, seriesDataU);
      newSeries = true;
    } 
    if (this.rssiSeriesDataMap.get("D-" + uuid) == null) {
      ObservableList<XYChart.Data<Date, Number>> seriesDataD = FXCollections.observableArrayList();
      this.rssiSeriesDataMap.put("D-" + uuid, seriesDataD);
      newSeries = true;
    } 
    if (sample.getUplinkRssi() != null) {
      addSample(sample.getUplinkTime(), sample.getUplinkRssi().intValue(), "U-" + uuid, this.rssiSeriesDataMap, this.rssiSeries);
      addSample(sample.getUplinkTime(), sample.getUplinkRssi().intValue(), "U-" + uuid, this.rssiSeriesDataMap, this.uplinkRssiSeries);
    } 
    addSample(sample.getUplinkTime(), sample.getRssi(), "D-" + uuid, this.rssiSeriesDataMap, this.rssiSeries);
    addSample(sample.getUplinkTime(), sample.getRssi(), "D-" + uuid, this.rssiSeriesDataMap, this.downlinkRssiSeries);
    if (newSeries)
      setUpRssiChart(); 
    return newSeries;
  }
  
  private void updateLatestSamples(PhySample sample) {
    String sampleUuid = sample.getUuid();
    Optional<PhySample> matchingObjects = this.latestPhySamples.stream().filter(p -> p.getUuid().equals(sampleUuid)).findAny();
    PhySample sampleInTable = matchingObjects.orElse(null);
    if (sampleInTable == null) {
      sampleInTable = new PhySample(sample.getUuid(), sample.getTimestamp(), sample.getUplinkRssi(), sample.getPackets(), sample.getRssi(), sample.getBatteryVoltage(), sample.getTemperature(), sample.getUplinkTime(), sample.getSampleTime());
      this.latestPhySamples.add(sampleInTable);
    } else {
      sampleInTable.setTimestamp(sample.getTimestamp());
      sampleInTable.setUplinkRssi(sample.getUplinkRssi());
      sampleInTable.setPackets(sample.getPackets());
      sampleInTable.setRssi(sample.getRssi());
      sampleInTable.setBatteryVoltage(sample.getBatteryVoltage());
      sampleInTable.setTemperature(sample.getTemperature());
      sampleInTable.setUplinkTime(sample.getUplinkTime());
      sampleInTable.setSampleTime(sample.getSampleTime());
    } 
    this.tableView.refresh();
  }
  
  public void cbUseOneChartChanged() {
    setUseOneChartForAllSeries(this.cbUseOneChart.isSelected());
  }
  
  public void setUseOneChartForAllSeries(boolean useOne) {
    if (useOne != this.useOneChartForAllSeries) {
      this.useOneChartForAllSeries = useOne;
      this.chartsContainer.getChildren().clear();
      if (this.useOneChartForAllSeries) {
        this.uplinkRssiChart = null;
        this.downlinkRssiChart = null;
      } else {
        this.rssiChart = null;
      } 
      setUpRssiChart();
    } 
  }
  
  public boolean getBSInfo() {return false;}
}
