package ubiik.wp.demo.controller;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.sql.Timestamp;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.Optional;
import javafx.application.Platform;
import javafx.collections.FXCollections;
import javafx.collections.ObservableList;
import javafx.fxml.FXML;
import javafx.scene.chart.LineChart;
import javafx.scene.chart.XYChart;
import javafx.scene.control.TableColumn;
import javafx.scene.control.TableView;
import javafx.scene.control.cell.PropertyValueFactory;
import ubiik.wp.controller.DevicesController;
import ubiik.wp.controller.MessageController;
import ubiik.wp.demo.model.TemperatureHumiditySample;
import ubiik.wp.model.messages.PSUplink;
import ubiik.wp.util.ApplicationLogger;

public class TemperatureViewController extends UplinksDemoViewController {
  private ObservableList<TemperatureHumiditySample> latestTempHumSamples = FXCollections.observableArrayList();
  
  private LinkedHashMap<String, ObservableList<XYChart.Data<Date, Number>>> temperatureSeriesDataMap;
  
  private LinkedHashMap<String, ObservableList<XYChart.Data<Date, Number>>> humiditySeriesMap;
  
  private ObservableList<XYChart.Series<Date, Number>> temperatureSeries;
  
  private ObservableList<XYChart.Series<Date, Number>> humiditySeries;
  
  private LineChart<Date, Number> temperatureChart;
  
  private LineChart<Date, Number> humidityChart;
  
  @FXML
  protected TableView<TemperatureHumiditySample> tableView;
  
  public TemperatureViewController(DevicesController devicesController, MessageController messageController) {
    super(devicesController, messageController);
    this.temperatureSeriesDataMap = new LinkedHashMap<>();
    this.humiditySeriesMap = new LinkedHashMap<>();
    this.temperatureSeries = FXCollections.observableArrayList();
    this.humiditySeries = FXCollections.observableArrayList();
  }
  
  public void initialize() {
    super.initialize();
    setUpTemperatureChart();
    setUpHumidityChart();
    setUpSamplesTable();
    setTitle("Temperature Demo");
    this.cbUseOneChart.setVisible(false);
  }
  
  private void setUpTemperatureChart() {
    this.temperatureChart = setUpLineChart(this.temperatureChart, this.temperatureSeries, "Temperature (ºC)");
  }
  
  private void setUpHumidityChart() {
    this.humidityChart = setUpLineChart(this.humidityChart, this.humiditySeries, "Humidity (%)");
  }
  
  private void setUpSamplesTable() {
    this.tableView.getColumns().clear();
    TableColumn<TemperatureHumiditySample, String> deviceCol = new TableColumn<>("End Device");
    TableColumn<TemperatureHumiditySample, Float> temperatureCol = new TableColumn<>("Temperature (ºC)");
    TableColumn<TemperatureHumiditySample, Integer> humidityCol = new TableColumn<>("Humidity");
    TableColumn<TemperatureHumiditySample, Timestamp> datelCol = new TableColumn<>("Date");
    deviceCol.prefWidthProperty().bind(this.tableView.widthProperty().multiply(0.35D));
    temperatureCol.prefWidthProperty().bind(this.tableView.widthProperty().multiply(0.2D));
    humidityCol.prefWidthProperty().bind(this.tableView.widthProperty().multiply(0.2D));
    datelCol.prefWidthProperty().bind(this.tableView.widthProperty().multiply(0.25D));
    deviceCol.setSortable(true);
    temperatureCol.setSortable(false);
    humidityCol.setSortable(false);
    datelCol.setSortable(true);
    deviceCol.setCellValueFactory(new PropertyValueFactory<>("uuid"));
    temperatureCol.setCellValueFactory(new PropertyValueFactory<>("temperature"));
    humidityCol.setCellValueFactory(new PropertyValueFactory<>("humidity"));
    datelCol.setCellValueFactory(new PropertyValueFactory<>("timestamp"));
    this.tableView.getColumns().addAll(new TableColumn[] { deviceCol, temperatureCol, humidityCol, datelCol });
    this.tableView.setItems(this.latestTempHumSamples);
  }
  
  private TemperatureHumiditySample getSampleFromUplink(PSUplink uplink) {
    TemperatureHumiditySample temperatureHumiditySample = null;
    byte[] data = uplink.getApplicationData();
    try {
      ByteBuffer buffer = ByteBuffer.wrap(data).order(ByteOrder.LITTLE_ENDIAN);
      if (buffer.get() == 1) {
        int temperature = buffer.getShort();
        int humidity = buffer.get();
        temperatureHumiditySample = new TemperatureHumiditySample(uplink.getTargetString(), uplink.getTimestamp(), temperature / 100, humidity);
        ApplicationLogger.debug("New sample: " + temperatureHumiditySample.toString());
      } 
    } catch (Exception exception) {}
    return temperatureHumiditySample;
  }
  
  protected void processUplink(final PSUplink uplink) {
    final TemperatureHumiditySample sample = getSampleFromUplink(uplink);
    if (sample != null)
      Platform.runLater(new Runnable() {
            public void run() {
              TemperatureViewController.this.addTemperatureSample(sample, uplink.getTargetString());
              TemperatureViewController.this.addHumiditySample(sample, uplink.getTargetString());
              TemperatureViewController.this.updateLatestSamples(sample);
            }
          }); 
  }
  
  private void addTemperatureSample(TemperatureHumiditySample sample, String uuid) {
    boolean newSeries = addSample(sample.getTimestamp(), (int)sample.getTemperature(), uuid, this.temperatureSeriesDataMap, this.temperatureSeries);
    if (newSeries)
      setUpTemperatureChart(); 
  }
  
  private void addHumiditySample(TemperatureHumiditySample sample, String uuid) {
    boolean newSeries = addSample(sample.getTimestamp(), sample.getHumidity(), uuid, this.humiditySeriesMap, this.humiditySeries);
    if (newSeries)
      setUpHumidityChart(); 
  }
  
  private void updateLatestSamples(TemperatureHumiditySample sample) {
    String sampleUuid = sample.getUuid();
    Optional<TemperatureHumiditySample> matchingObjects = this.latestTempHumSamples.stream().filter(p -> p.getUuid().equals(sampleUuid)).findAny();
    TemperatureHumiditySample oldSample = matchingObjects.orElse(null);
    if (oldSample != null)
      this.latestTempHumSamples.remove(oldSample); 
    this.latestTempHumSamples.add(sample);
  }
  
  public void cbUseOneChartChanged() {
    setUpTemperatureChart();
    setUpHumidityChart();
  }
  
  public boolean getBSInfo() {return false;}
}
