package ubiik.wp.model.devices;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import ubiik.wp.controller.DevicesObserver;
import ubiik.wp.controller.EndDeviceNameAndIdxManager;
import ubiik.wp.controller.UserManager;
import ubiik.wp.model.bsparameters.BSParameter;
import ubiik.wp.model.bsparameters.BSParameterFactory;
import ubiik.wp.model.fpga.FPGARegisterFactory;
import ubiik.wp.model.messages.PSDownlink;
import ubiik.wp.model.messages.PSPong;
import ubiik.wp.model.messages.PSUplink;
import ubiik.wp.model.sdr.SDRParameter;
import ubiik.wp.model.sdr.SDRParameterFactory;
import ubiik.wp.view.DisplayConstants;

public class DevicesStorage {
  private UUBaseStation baseStation;
  
  private LinkedHashMap<String, UUEndDevice> devices;
  
  private Map<Integer, MulticastGroup> multicastGroups;
  
  private List<DevicesObserver> observers;
  
  private LinkedHashMap<Integer, SDRParameter> sdrParameters;
  
  private UUSIB0 sib0;
  
  private int edIdxCounter;
  
  private Map<String, Integer> devicesIndexes;
  
  public static final byte DEVICE_TYPE_ED = 0;
  
  public static final byte DEVICE_TYPE_FOREIGN_ED = 100;
  
  private DevicesStorage() {
    this.observers = new LinkedList<>();
    setUp();
  }
  
  public synchronized void restart() {
    setUp();
    notifyObserversDevicesCleared();
  }
  
  private synchronized void setUp() {
    if (this.baseStation == null) {
      this.baseStation = new UUBaseStation();
    } else {
      this.baseStation.restart();
    } 
    FPGARegisterFactory factory = UserManager.getInstance().getFpgaRegisterFactory();
    this.baseStation.setRegisters(factory.loadRegisters());
    BSParameterFactory bsParamFactory = UserManager.getInstance().getBsParamFactory();
    this.baseStation.setBsParameters(bsParamFactory.loadBSParameters("parameters"));
    this.devices = new LinkedHashMap<>();
    this.multicastGroups = new HashMap<>();
    this.sdrParameters = new LinkedHashMap<>();
    this.sib0 = new UUSIB0((short)0, (short)0, (short)0, (short)0, null);
    SDRParameterFactory sdrFactory = UserManager.getInstance().getSdrParameterFactory();
    setSDRParameters(sdrFactory.loadSDRParameters());
    this.edIdxCounter = 0;
    this.devicesIndexes = new HashMap<>();
  }
  
  public synchronized void clearDevices() {
    this.devices.clear();
  }
  
  private static class Holder {
    static final DevicesStorage INSTANCE = new DevicesStorage();
  }
  
  public static DevicesStorage getInstance() {
    return Holder.INSTANCE;
  }
  
  public synchronized UUBaseStation getBaseStation() {
    return this.baseStation;
  }
  
  public synchronized byte[] getBsid() {
    return this.baseStation.getBsid();
  }
  
  public synchronized void setBsid(byte[] bsid) {
    this.baseStation.setBsid(bsid);
    notifyObservers();
  }
  
  public synchronized Collection<UUEndDevice> getDevices() {
    Collection<UUEndDevice> eds = new LinkedList<>(this.devices.values());
    return eds;
  }
  
  public synchronized UUEndDevice getDeviceForId(byte[] uuid) {
    return this.devices.get(UUEndDevice.getStringForUUIDString(uuid));
  }
  
  public synchronized boolean deviceExists(byte[] uuid) {
    return (getDeviceForId(uuid) != null);
  }
  
  private int getIdxForEDtoAdd(byte[] uuid) {
    String strUuid = UUEndDevice.getStringForUUIDString(uuid);
    Integer idx = this.devicesIndexes.get(strUuid);
    if (idx == null) {
      idx = Integer.valueOf(this.edIdxCounter++);
      this.devicesIndexes.put(strUuid, idx);
    } 
    return idx.intValue();
  }
  
  public void setIdxForEds(Map<String, Integer> edsIndx, int startIdx) {
    int idxOffset = (startIdx >= 0) ? (startIdx + 1) : 0;
    for (Map.Entry<String, Integer> entry : this.devicesIndexes.entrySet()) {
      if (!edsIndx.containsKey(entry.getKey())) {
        entry.setValue(Integer.valueOf(idxOffset));
        idxOffset++;
      } 
    } 
    if (idxOffset >= 0)
      this.edIdxCounter = idxOffset; 
    this.devicesIndexes.putAll(edsIndx);
  }
  
  public synchronized UUEndDevice addDevice(byte[] uuid) {
    if (getDeviceForId(uuid) == null) {
      UUEndDevice endDevice = new UUEndDevice(uuid, getIdxForEDtoAdd(uuid));
      this.devices.put(UUEndDevice.getStringForUUIDString(uuid), endDevice);
      String name = EndDeviceNameAndIdxManager.getInstance().getNameForUUID(endDevice.getUUIDString());
      if (name != null)
        endDevice.setName(name); 
      notifyObserversDeviceAddded(endDevice);
      return endDevice;
    } 
    return null;
  }
  
  public synchronized void setEndDeviceStatus(byte[] uuid, byte status, Date timeStamp, short rssi, Date lastActivity, Date dateConnected, Long totalUplinksUNACK, Long totalUplinksACK, byte deviceType, byte[] firmwareVersion) {
	    UUEndDevice device = getDeviceForId(uuid);
	    if (device != null) {
	      device.setStatus(status);
	      if (timeStamp != null)
	    	  device.setDateAdded(timeStamp); 
	      if (rssi != Short.MAX_VALUE)
	    	  device.setRegistrationRssi(Short.valueOf(rssi)); 
	      if (lastActivity != null)
	    	  device.setLastActivity(lastActivity); 
	      if (dateConnected != null)
	    	  device.setDateConnected(dateConnected); 
	      if (totalUplinksACK != null)
	    	  device.setTotalUlsACK(totalUplinksACK.longValue()); 
	      if (totalUplinksUNACK != null)
	    	  device.setTotalUlsUNACK(totalUplinksUNACK.longValue());
	      if (firmwareVersion != null)
	    	  device.setFirmwareVersion(firmwareVersion);
	    	  
		  if (deviceType == DEVICE_TYPE_ED)
	    	  device.setEndDeviceType("ED");
	      else if (deviceType == DEVICE_TYPE_FOREIGN_ED)
	    	  device.setEndDeviceType("ForeignED");	  
		  
	      notifyObservers();
	    } 
	  }
  
  public synchronized void setEndDeviceStatus(byte[] uuid, byte status, Date timeStamp, short rssi, Date lastActivity, Date dateConnected, Long totalUplinksUNACK, Long totalUplinksACK, byte deviceType, byte[] firmwareVersion, Long joinFoi) {
    UUEndDevice device = getDeviceForId(uuid);
    if (device != null) {
      device.setStatus(status);
      if (timeStamp != null)
    	  device.setDateAdded(timeStamp); 
      if (rssi != Short.MAX_VALUE)
    	  device.setRegistrationRssi(Short.valueOf(rssi)); 
      if (lastActivity != null)
    	  device.setLastActivity(lastActivity); 
      if (dateConnected != null)
    	  device.setDateConnected(dateConnected); 
      if (totalUplinksACK != null)
    	  device.setTotalUlsACK(totalUplinksACK.longValue()); 
      if (totalUplinksUNACK != null)
    	  device.setTotalUlsUNACK(totalUplinksUNACK.longValue());
      if (firmwareVersion != null)
    	  device.setFirmwareVersion(firmwareVersion);
      if (joinFoi != null)
    	  device.setJoinFoi(joinFoi);
    	  
	  if (deviceType == DEVICE_TYPE_ED)
    	  device.setEndDeviceType("ED");
      else if (deviceType == DEVICE_TYPE_FOREIGN_ED)
    	  device.setEndDeviceType("ForeignED");	  
	  
      notifyObservers();
    } 
  }
  
  public void removeEndDevice(UUEndDevice endDevice) {
    this.devices.remove(new String(endDevice.getUuid()));
    notifyObserversDeviceRemoved(endDevice);
  }
  
  public synchronized boolean setScheduleTransferForDevice(byte[] uuid, byte flag, int window, int interval) {
    boolean valid = false;
    UUEndDevice device = getDeviceForId(uuid);
    if (device != null) {
      valid = device.setScheduleTransferConfig(flag, window, interval);
      if (valid)
        notifyObservers(); 
    } 
    return valid;
  }
  
  public synchronized void addUplinkToDevice(PSUplink uplink) {
    UUEndDevice device = getDeviceForId(uplink.getTarget());
    if (device != null) {
      device.incrementUplinks(uplink);
      uplink.setEndDevice(device);
      notifyObservers();
    } 
  }
  
  public synchronized void addDownlinkToDevice(PSDownlink downlink) {
    UUEndDevice device = getDeviceForId(downlink.getTarget());
    if (device != null) {
      device.incrementDownlinks(downlink);
      downlink.setEndDevice(device);
      notifyObservers();
    } 
  }
  
  public synchronized void updateDevicesInMulticastGroup(int groupId, List<byte[]> devices, List<Byte> status) {
    if (status != null && 
      status.size() != devices.size())
      throw new IllegalArgumentException(); 
    MulticastGroup mcg = null;
    if (!this.multicastGroups.containsKey(Integer.valueOf(groupId))) {
      mcg = new MulticastGroup((short)groupId);
      this.multicastGroups.put(Integer.valueOf(groupId), mcg);
    } else {
      mcg = this.multicastGroups.get(Integer.valueOf(groupId));
    } 
    byte[] uuid = null;
    for (int i = 0; i < devices.size(); i++) {
      uuid = devices.get(i);
      UUEndDevice ed = getDeviceForId(uuid);
      if (ed == null)
        ed = addDevice(uuid); 
      if (status != null) {
        if (((Byte)status.get(i)).byteValue() == 4) {
          mcg.removeED(ed);
        } else {
          mcg.addOrUpdateED(ed, ((Byte)status.get(i)).byteValue());
        } 
      } else {
        mcg.addED(ed);
      } 
    } 
    notifyObservers();
  }
  
  public synchronized void removeDevicesFromMulticastGrouo(int groupId, List<byte[]> devices) {
    if (this.multicastGroups.containsKey(Integer.valueOf(groupId))) {
      MulticastGroup mcg = this.multicastGroups.get(Integer.valueOf(groupId));
      for (byte[] uuid : devices) {
        UUEndDevice ed = getDeviceForId(uuid);
        if (ed == null)
          mcg.removeED(ed); 
      } 
      notifyObservers();
    } 
  }
  
  public synchronized void setMulticastGroups(Map<Integer, List<byte[]>> multicastGroups, Map<Integer, List<Byte>> status) {
    clearMulticastGroups();
    addMulticastGroups(multicastGroups, status);
  }
  
  public synchronized void addMulticastGroups(Map<Integer, List<byte[]>> multicastGroups, Map<Integer, List<Byte>> status) {
    for (Integer groupId : multicastGroups.keySet())
      updateDevicesInMulticastGroup(groupId.intValue(), multicastGroups.get(groupId), status.get(groupId)); 
    notifyObservers();
  }
  
  public synchronized boolean addMulticastGroup(MulticastGroup mcg) {
    boolean groupExists = mutlicastGroupExists(mcg.getGroupId());
    if (!groupExists)
      this.multicastGroups.put(Integer.valueOf(mcg.getGroupId()), mcg); 
    return !groupExists;
  }
  
  public synchronized Map<Integer, MulticastGroup> getMulticastGroups() {
    return this.multicastGroups;
  }
  
  public synchronized boolean mutlicastGroupExists(int groupId) {
    return this.multicastGroups.containsKey(Integer.valueOf(groupId));
  }
  
  public synchronized void clearMulticastGroups() {
    this.multicastGroups.clear();
  }
  
  private synchronized void notifyObservers() {
    for (DevicesObserver observer : this.observers)
      observer.updateDevices(); 
  }
  
  private synchronized void notifyObserversDevicesCleared() {
    for (DevicesObserver observer : this.observers)
      observer.devicesCleared(); 
  }
  
  private synchronized void notifyObserversDeviceAddded(UUEndDevice endDevice) {
    for (DevicesObserver observer : this.observers)
      observer.deviceAdded(endDevice); 
  }
  
  private synchronized void notifyObserversDeviceRemoved(UUEndDevice endDevice) {
    for (DevicesObserver observer : this.observers)
      observer.deviceRemoved(endDevice); 
  }
  
  public synchronized void addObdserver(DevicesObserver ovserver) {
    this.observers.add(ovserver);
  }
  
  public synchronized void removeObdserver(DevicesObserver ovserver) {
    this.observers.remove(ovserver);
  }
  
  public synchronized UUSIB0 getSib() {
    return this.sib0;
  }
  
  public synchronized void setSib(UUSIB0 sib) {
    this.sib0 = sib;
    notifyObservers();
  }
  
  public synchronized void setBSRParameterValue(int paramIdx, Long value) {
    BSParameter param = this.baseStation.getBsParameters().get(paramIdx);
    if (param != null) {
      param.setValue(value);
      notifyObservers();
    } 
  }
  
  public synchronized ArrayList<BSParameter> getBSParameters() {
    return this.baseStation.getBsParameters();
  }
  
  public synchronized BSParameter getBSRParameterForIdx(int paramIdx) {
    BSParameter parameter = null;
    parameter = this.baseStation.getBsParameterAtIdx(paramIdx);
    return parameter;
  }
  
  public synchronized void setSDRParameters(ArrayList<SDRParameter> sdrParameters) {
    for (SDRParameter sdrParameter : sdrParameters)
      this.sdrParameters.put(Integer.valueOf(sdrParameter.getId()), sdrParameter); 
    notifyObservers();
  }
  
  public synchronized void setSDRParameterValue(int sdrParaId, String value) {
    SDRParameter param = this.sdrParameters.get(Integer.valueOf(sdrParaId));
    if (param != null) {
      param.setValuFromString(value);
      notifyObservers();
    } 
  }
  
  public synchronized Collection<SDRParameter> getSDRParameters() {
    return this.sdrParameters.values();
  }
  
  public synchronized SDRParameter getSDRParameterForId(int paramId) {
    SDRParameter parameter = null;
    parameter = this.sdrParameters.get(Integer.valueOf(paramId));
    return parameter;
  }
  
  public synchronized Integer getRunMode() {
    return this.baseStation.getRunMode();
  }
  
  public synchronized void setRunMode(Integer runMode) {
    this.baseStation.setRunMode(runMode);
    notifyObservers();
  }
  
  public synchronized byte[] getBsVersion() {
    return this.baseStation.getVersion();
  }
  
  public synchronized void setBsVersion(byte[] version) {
    this.baseStation.setVersion(version);
  }
  
  public synchronized void setBsBuildDate(Date date) {
    this.baseStation.setBuildDate(date);
  }
  
  public synchronized Date getBsBuildDate() {
    return this.baseStation.getBuildDate();
  }
  
  public synchronized void setBsVersionAndBuildDate(byte[] version, String strDate) {
    if (version != null)
      setBsVersion(version); 
    if (strDate != null)
      try {
        Date date = DisplayConstants.SIMPLE_DATE_FORMAT_YYYYMMDDTHHMMSS.parse(strDate.trim());
        setBsBuildDate(date);
      } catch (Exception e) {
        e.printStackTrace();
      }  
  }
  
  public synchronized PSPong getLastPongForDevice(byte[] uueid) {
    PSPong pong = null;
    UUEndDevice ed = getDeviceForId(uueid);
    if (ed != null)
      pong = ed.getLastPong(); 
    return pong;
  }
  
  public synchronized void setLastPongForDevice(PSPong lastPong) {
    UUEndDevice ed = getDeviceForId(lastPong.getUueid());
    if (ed != null)
      ed.setLastPong(lastPong); 
  }
  
  public synchronized void setNumEndDevices (int numEDs) {
	  this.baseStation.setNumEndDevices(numEDs);
  }
  
  public synchronized int getNumEndDevices () {
	  return this.baseStation.getNumEndDevices();
  }
}
