package ubiik.wp.model.devices;

import java.util.Arrays;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import ubiik.utils.DatatypeConverter;
import ubiik.wp.model.devices.mcs.MCSSetting;
import ubiik.wp.model.messages.PSDownlink;
import ubiik.wp.model.messages.PSPong;
import ubiik.wp.model.messages.PSUplink;

public class UUEndDevice {
  public static final short DEVICE_NO_RSSI = 32767;
  
  public static final int MAX_MTU = 4096;
  
  public static final int MIN_MTU = 0;
  
  public static final byte STATUS_NONE = 0;
  
  public static final byte STATUS_REGISTERING = 1;
  
  public static final byte STATUS_CONNECTED = 2;
  
  public static final byte STATUS_DISCONNECTED = 3;
  
  public static final byte STATUS_NOT_PROVISIONED = 4;
  
  public static final byte STATUS_NOT_FOUND = 5;
  
  public static final byte STATUS_REGISTERED = 6;
  
  public static final byte STATUS_REJECTED = 7;
  
  public static final byte STATUS_INACTIVE = -2;
  
  public static final byte STATUS_UNDEFINED = -1;
  
  private static final Map<Byte, String> STATUS_NAMES = new HashMap<>();
  
  private final byte[] uuid;
  
  private Date dateAdded;
  
  private Date dateConnected;
  
  private long uplinksCounter;
  
  private long downlinksCounter;
  
  private Date lastUplink;
  
  private Date lastDownlink;
  
  private Date lastActivity;
  
  private byte status;
  
  private String name;
  
  private long totalUlsACK;
  
  private long totalUlsUNACK;
  
  private ScheduleTransferConfig scheduleTransferConfig;
  
  private int idx;
  
  private Short registrationRssi;
  
  private Short lastUplinkRssi;
  
  private boolean userCreated;
  
  private MCSSetting mcsSettings;
  
  private PSPong lastPong;
  
  private String endDeviceType;
  
  private String firmwareVersion;
  
  private Long joinFoi;
  
  static {
    STATUS_NAMES.put(Byte.valueOf((byte)0), "Unknown");
    STATUS_NAMES.put(Byte.valueOf((byte)1), "Registering");
    STATUS_NAMES.put(Byte.valueOf((byte)2), "Connected");
    STATUS_NAMES.put(Byte.valueOf((byte)3), "Disconnected");
    STATUS_NAMES.put(Byte.valueOf((byte)4), "Not Provisioned");
    STATUS_NAMES.put(Byte.valueOf((byte)5), "Not Found");
    STATUS_NAMES.put(Byte.valueOf((byte)6), "Registered");
    STATUS_NAMES.put(Byte.valueOf((byte)7), "Rejected");
    STATUS_NAMES.put(Byte.valueOf((byte)-2), "Innactive");
  }
  
  public UUEndDevice(byte[] uuid, int idx) {
    this.uuid = uuid;
    this.dateAdded = null;
    this.uplinksCounter = 0L;
    this.downlinksCounter = 0L;
    this.totalUlsACK = 0L;
    this.totalUlsUNACK = 0L;
    this.lastUplink = null;
    this.lastDownlink = null;
    this.status = 0;
    this.idx = idx;
    this.registrationRssi = null;
    this.lastUplinkRssi = null;
    this.name = getUUIDString();
    setUserCreated(false);
    this.lastActivity = null;
    this.mcsSettings = null;
    this.endDeviceType = null;
    this.firmwareVersion = null;
    this.joinFoi = null;
  }
  
  public int hashCode() {
    return Arrays.hashCode(this.uuid);
  }
  
  public boolean equals(Object obj) {
    if (obj == null)
      return false; 
    if (!UUEndDevice.class.isAssignableFrom(obj.getClass()))
      return false; 
    UUEndDevice other = (UUEndDevice)obj;
    return Arrays.equals(this.uuid, other.getUuid());
  }
  
  public synchronized void incrementUplinks(PSUplink uplink) {
    this.uplinksCounter++;
    this.lastUplink = uplink.getCreationDate();
    this.lastActivity = Calendar.getInstance().getTime();
    if (uplink.getRssi() != null)
      this.lastUplinkRssi = Short.valueOf(uplink.getRssi().shortValue()); 
  }
  
  public synchronized void incrementDownlinks(PSDownlink downlink) {
    this.downlinksCounter++;
    this.lastDownlink = downlink.getTimestamp();
  }
  
  public synchronized byte[] getUuid() {
    return this.uuid;
  }
  
  public synchronized Date getDateAdded() {
    return this.dateAdded;
  }
  
  public synchronized void setDateAdded(Date dateAdded) {
    this.dateAdded = dateAdded;
  }
  
  public synchronized Date getLastUplinkDate() {
    return this.lastUplink;
  }
  
  public synchronized Date getLastDownlinkDate() {
    return this.lastDownlink;
  }
  
  public synchronized long getUplinksCounter() {
    return this.uplinksCounter;
  }
  
  public synchronized void setUplinksCounter(long uplinksCounter) {
    this.uplinksCounter = uplinksCounter;
  }
  
  public synchronized long getDownlinksCounter() {
    return this.downlinksCounter;
  }
  
  public synchronized String getUUIDString() {
    //return DatatypeConverter.printHexBinary(this.uuid);
    //Display IMSI only : index: 17 ~ 31(15 digits IMSI), 0 ~ 16 (17 digits) are paddings for UUID
    return DatatypeConverter.printHexBinary(this.uuid).substring(17);
  }
  
  public static final synchronized String getStringForUUIDString(byte[] uuid) {
    //return DatatypeConverter.printHexBinary(uuid);
    //Display IMSI only : index: 17 ~ 31(15 digits IMSI), 0 ~ 16 (17 digits) are paddings for UUID
    return DatatypeConverter.printHexBinary(uuid).substring(17);
  }
  
  public synchronized String getStatusString() {
    return getStringForStatus(Byte.valueOf(this.status));
  }
  
  public static final String getStringForStatus(Byte status) {
    String statusStr = STATUS_NAMES.get(status);
    if (statusStr == null)
      statusStr = Integer.toString(0xFF & status.byteValue()); 
    return statusStr;
  }
  
  public synchronized byte getStatus() {
    return this.status;
  }
  
  public synchronized void setStatus(byte status) {
    this.status = status;
    if (status != -2)
      this.lastActivity = Calendar.getInstance().getTime(); 
  }
  
  public synchronized ScheduleTransferConfig getScheduleTransferConfig() {
    return this.scheduleTransferConfig;
  }
  
  public synchronized boolean setScheduleTransferConfig(byte flag, int window, int interval) {
    boolean valid = false;
    ScheduleTransferConfig sct = new ScheduleTransferConfig();
    valid = sct.setStatus(flag, window, interval);
    if (valid)
      this.scheduleTransferConfig = sct; 
    return valid;
  }
  
  public synchronized int getIdx() {
    return this.idx;
  }
  
  public synchronized Short getRegistrationRssi() {
    return this.registrationRssi;
  }
  
  public synchronized void setRegistrationRssi(Short registrationRssi) {
    this.registrationRssi = registrationRssi;
  }
  
  public synchronized Short getLastUplinkRssi() {
    return this.lastUplinkRssi;
  }
  
  public synchronized void setLastUplinkRssi(Short lastUplinkRssi) {
    this.lastUplinkRssi = lastUplinkRssi;
  }
  
  public synchronized String getName() {
    return this.name;
  }
  
  public synchronized void setName(String name) {
    this.name = name;
  }
  
  public synchronized String toString() {
    if (getName().compareTo(getUUIDString()) == 0)
      return getName(); 
    return getName() + " (" + getName() + ")";
  }
  
  public synchronized boolean isUserCreated() {
    return this.userCreated;
  }
  
  public synchronized void setUserCreated(boolean userCreated) {
    this.userCreated = userCreated;
  }
  
  public synchronized Date getLastActivity() {
    return this.lastActivity;
  }
  
  public synchronized void setLastActivity(Date lastAct) {
    this.lastActivity = lastAct;
  }
  
  public synchronized Date getDateConnected() {
    return this.dateConnected;
  }
  
  public synchronized void setDateConnected(Date dateConnected) {
    this.dateConnected = dateConnected;
  }
  
  public synchronized long getTotalUlsACK() {
    return this.totalUlsACK;
  }
  
  public synchronized void setTotalUlsACK(long totalUlsACK) {
    this.totalUlsACK = totalUlsACK;
  }
  
  public synchronized long getTotalUlsUNACK() {
    return this.totalUlsUNACK;
  }
  
  public synchronized void setTotalUlsUNACK(long totalUlsUNACK) {
    this.totalUlsUNACK = totalUlsUNACK;
  }
  
  public synchronized MCSSetting getMCSSetting() {
    return this.mcsSettings;
  }
  
  public synchronized void setMCSFlags(short uplinkMCSFlags, short downlinkMCSFlags) {
    if (this.mcsSettings == null)
      this.mcsSettings = new MCSSetting(); 
    this.mcsSettings.setDownlinkMcsFlags(downlinkMCSFlags);
    this.mcsSettings.setUplinkMcsFlags(uplinkMCSFlags);
  }
  
  public synchronized void setMCSValues(byte downlinkMCSValue, byte uplinkMCSValue, byte uplinkMCSValueNB) {
    if (this.mcsSettings == null)
      this.mcsSettings = new MCSSetting(); 
    this.mcsSettings.setCurrentDownlinkMcsValue(downlinkMCSValue);
    this.mcsSettings.setCurrentuplinkMcsValue(uplinkMCSValue);
    this.mcsSettings.setCurrentuplinkMcsValueNB(uplinkMCSValueNB);
  }
  
  public synchronized PSPong getLastPong() {
    return this.lastPong;
  }
  
  public synchronized void setLastPong(PSPong lastPong) {
    this.lastPong = lastPong;
  }
  
  public synchronized void setEndDeviceType(String endDeviceType) {
	  this.endDeviceType = endDeviceType;
  }
  
  public synchronized String getEndDeviceType() {
	  return this.endDeviceType;
  }
  
  public synchronized String getFirmwareVersion() {
	  return this.firmwareVersion;
  }
  
  public synchronized void setFirmwareVersion(byte[] firmwareVersion) {
	  this.firmwareVersion = "";
	  for (	int i = 0 ; i < firmwareVersion.length; i++	) {
		  if (this.firmwareVersion.length() != 0)
			  this.firmwareVersion = this.firmwareVersion + ".";
		  this.firmwareVersion = this.firmwareVersion + firmwareVersion[i];
	  } 
  }
  
  public synchronized void setJoinFoi (long joinFoi) {
	  this.joinFoi = joinFoi;
  }
  
  public synchronized Long getJoinFoi () {
	  return this.joinFoi;
  }
  
}
