package ubiik.wp.model.messages;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Locale;
import ubiik.utils.DatatypeConverter;

public abstract class PSMessage {
  public static final int NO_APPLICATION_ID = -1;
  
  public static final int PROTOCOL_VERSION_UNKNOWN = -1;
  
  public static final int BYTES_FOR_SIZE = 4;
  
  protected int messageId;
  
  protected long applicationId;
  
  protected byte[] data;
  
  private byte typeCode;
  
  private Date timestamp;
  
  private static int messageIdCounter = 0;
  
  private static Object counterLock = new Object();
  
  private PSMessageResponse response;
  
  protected int protocolVersion;
  
  public static final byte PSMESSAGE_TYPE_AUTHENTICATED = 0;
  
  public static final byte PSMESSAGE_TYPE_UPLINK = 1;
  
  public static final byte PSMESSAGE_TYPE_DOWNLINK = 2;
  
  public static final byte PSMESSAGE_TYPE_MULTICAST_DOWNLINK = 3;
  
  public static final byte PSMESSAGE_TYPE_COMMAND = 4;
  
  public static final byte PSMESSAGE_TYPE_DEVICE_STATUS = 5;
  
  public static final byte PSMESSAGE_TYPE_ACK = 6;
  
  public static final byte PSMESSAGE_TYPE_LOG = 7;
  
  public static final byte PSMESSAGE_TYPE_DOWNLINK_ACK = 8;
  
  public static final byte PSMESSAGE_TYPE_JSON = 9;
  
  public static final byte PSMESSAGE_TYPE_FULL_UPLINK = 10;
  
  public static final byte PSMESSAGE_TYPE_NAK = 11;
  
  public static final byte PSMESSAGE_TYPE_MULTICAST_NOTIFICATION = 12;
  
  public static final byte PSMESSAGE_TYPE_PONG = 13;
  
  public static final byte PSMESSAGE_TYPE_NEW_DISCOVERY = 124;

  public static final byte PSMESSAGE_TYPE_DISCOVERY = 127;
  
  public static final byte PSMESSAGE_TYPE_DOWNLINK_RESP = -126;
  
  public static final byte PSMESSAGE_TYPE_MULTICAST_DOWNLINK_RESP = -125;
  
  public static final byte PSMESSAGE_TYPE_COMMAND_RESP = -124;
  
  public static final byte PSMESSAGE_TYPE_HW_INFO = 14;
  
  public static final byte PSMESSAGE_TYPE_COMMAND_UNDEFINED = -1;
  
  public PSMessage() {
    this(-1, new byte[0], (byte)-1);
  }
  
  public PSMessage(int messageId, int applicationID, byte[] messageData, byte typeCode) {
    this.messageId = messageId;
    this.applicationId = applicationID;
    this.data = messageData;
    this.typeCode = typeCode;
    this.timestamp = Calendar.getInstance().getTime();
    this.response = null;
    this.protocolVersion = -1;
  }
  
  public PSMessage(int applicationID, byte[] messageData, byte typeCode) {
    this(getNextMessageId(), applicationID, messageData, typeCode);
  }
  
  private static int getNextMessageId() {
    int nextMid = -1;
    synchronized (counterLock) {
      nextMid = messageIdCounter;
      messageIdCounter++;
    } 
    return nextMid;
  }
  
  protected abstract byte[] getMessageDataBytes() throws MessageFormatException;
  
  public byte[] toBytes() {
    byte[] mHeader = createMessageHeader();
    byte[] dataBytes = getMessageDataBytes();
    ByteBuffer buffer = ByteBuffer.allocate(mHeader.length + dataBytes.length);
    buffer.order(ByteOrder.LITTLE_ENDIAN);
    buffer.put(mHeader);
    buffer.put(dataBytes);
    return buffer.array();
  }
  
  protected byte[] createMessageHeader() {
    short mID = (short)(this.messageId & 0xFFFF);
    byte mType = (byte)(getTypeCode() & 0xFF);
    ByteBuffer buffer = ByteBuffer.allocate(3);
    buffer.order(ByteOrder.LITTLE_ENDIAN);
    buffer.putShort(mID);
    buffer.put(mType);
    return buffer.array();
  }
  
  public byte getTypeCode() {
    return this.typeCode;
  }
  
  protected void setTypeCode(byte typeCode) {
    this.typeCode = typeCode;
  }
  
  public int getMessageId() {
    return this.messageId;
  }
  
  
  public Date getTimestamp() {
    return this.timestamp;
  }
  
  public String getTimestampStr() {
    SimpleDateFormat sdf = new SimpleDateFormat("HH:mm:ss yyyy-MM-dd", Locale.ENGLISH);
    return sdf.format(this.timestamp);
  }
  
  public String getDataString() {
    return DatatypeConverter.printHexBinary(getMessageDataBytes());
  }
  
  public long getApplicationId() {
    return this.applicationId;
  }
  
  public byte[] getApplicationData() {
    return this.data;
  }
  
  public int getApplicationDataSize() {
    return this.data.length;
  }
  
  public String getApplicationDataString() {
    return DatatypeConverter.printHexBinary(getApplicationData());
  }
  
  public synchronized PSMessageResponse getMessageResponse() {
    return this.response;
  }
  
  public synchronized void setResponse(PSMessageResponse response) {
    this.response = response;
  }
  
  public int getProtocolVersion() {
    return this.protocolVersion;
  }
}
