package ubiik.wp.model.messages;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;
import ubiik.wp.util.ApplicationLogger;

public final class PSUplink extends PSLink {
  private static final int BYTES_FOR_UPLINK_ID = 2;
  
  private static final int BYTES_FOR_BASE_ID = 2;
  
  private static final int BYTES_FOR_ENDPOINT = 1;
  
  private static final int BYTES_FOR_HEADER_SIZE = 4;
  
  private static final int BYTES_FOR_PAYLOAD_SIZE = 4;
  
  private short baseId;
  
  private Integer rssi;
  
  private Date creationDate;
  
  private Long foi;
  
  public PSUplink(int uplinkId, byte[] messageData, byte[] target, short baseid) {
    super(uplinkId, (byte)1, messageData, target);
    this.baseId = baseid;
    this.rssi = null;
  }
  
  public PSUplink(int uplinkId, byte[] messageData, byte[] target, int endpoint, short baseid) {
    super(uplinkId, (byte)1, messageData, target, endpoint);
    this.baseId = baseid;
    this.rssi = null;
  }
  
  public PSUplink(byte[] data, int messageId, int messageType) {
    this.messageId = messageId;
    this.target = new byte[16];
    if (messageType == 1) {
      createUplinkV1(data, messageId);
    } else if (messageType == 10) {
      createUplinkV1_0_8(data, messageId);
    } else {
      throw new IllegalArgumentException("Invalid Message Type for Uplink");
    } 
    ApplicationLogger.debug("[" + getApplicationId() + "] [UL TIME " + getCreationDateStr() + "]");
  }
  
  private void createUplinkV1(byte[] data, int messageId) {
    try {
      ByteBuffer buffer = ByteBuffer.wrap(data).order(ByteOrder.LITTLE_ENDIAN);
      this.applicationId = (0xFFFF & buffer.getShort());
      this.baseId = buffer.getShort();
      this.endpoint = 0xFF & buffer.get();
      buffer.get(this.target);
      this.data = new byte[buffer.capacity() - buffer.position()];
      buffer.get(this.data);
      this.creationDate = new Date(getTimestamp().getTime());
      this.rssi = null;
      this.protocolVersion = -1;
    } catch (Exception e) {
      throw new IllegalArgumentException();
    } 
  }
  
  private void createUplinkV1_0_8(byte[] data, int messageId) {
    try {
      ByteBuffer buffer = ByteBuffer.wrap(data).order(ByteOrder.LITTLE_ENDIAN);
      int headerSize = buffer.getInt();
      int payloadSize = buffer.getInt();
      this.protocolVersion = buffer.getInt();
      this.applicationId = buffer.getInt();
      long timeCreation = (0xFFFFFFFFL & buffer.getInt()) * 1000L;
      this.creationDate = new Date(timeCreation);
      this.baseId = buffer.getShort();
      this.endpoint = 0xFFFF & buffer.getShort();
      this.rssi = Integer.valueOf(buffer.getShort());
      buffer.get(this.target);
      this.foi = Long.valueOf(buffer.getInt());
      this.data = new byte[payloadSize];
      buffer.position(headerSize + 4 + 4);
      buffer.get(this.data);
    } catch (Exception e) {
      e.printStackTrace();
      throw new IllegalArgumentException();
    } 
  }
  
  protected byte[] getMessageDataBytes() throws MessageFormatException {
    if (this.target.length != 16)
      throw new MessageFormatException(); 
    ByteBuffer buffer = ByteBuffer.allocate(21 + this.data.length).order(ByteOrder.LITTLE_ENDIAN);
    buffer.putShort((short)(int)(this.applicationId & 0xFFFFL));
    buffer.putShort(this.baseId);
    buffer.put((byte)(this.endpoint & 0xFF));
    buffer.put(this.target);
    buffer.put(this.data);
    return buffer.array();
  }
  
  public Integer getRssi() {
    return this.rssi;
  }
  
  public String getRssiStr() {
    if (this.rssi != null)
      return Integer.toString(this.rssi.intValue()); 
    return "-";
  }
  
  public Date getCreationDate() {
    return this.creationDate;
  }
  
  public String getCreationDateStr() {
    SimpleDateFormat sdf = new SimpleDateFormat("HH:mm:ss yyyy-MM-dd", Locale.ENGLISH);
    return sdf.format(this.creationDate);
  }
  
  public Long getFoi() {
	  return this.foi;
  }
  
  public String toString() {
    return Long.toString(this.applicationId) + "\t" + Long.toString(this.applicationId) + "\t" + getTargetString() + "\t" + 
      getApplicationDataString() + "\t" + getApplicationDataSize() + "\t" + 
      getRssiStr() + "\t" + getEndpointStr();
  }
}
