package ubiik.wp.model.messages.pscommands;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import ubiik.wp.model.messages.MessageFormatException;
import ubiik.wp.model.messages.PSMessage;
import ubiik.wp.model.messages.pscommands.responses.PSCommandResponse;

public abstract class PSCommand extends PSMessage {
  protected short commandType = -1;
  
  private PSCommandResponse response;
  
  private PSCommandDelegate delegate;
  
  private static int commandIdCounter = 0;
  
  private static Object commandCounterLock = new Object();
  
  public static final short COMMAND_TYPE_GET_BASE_CHANNEL = 1;
  
  public static final short COMMAND_TYPE_SET_BASE_CHANNEL = 2;
  
  public static final short COMMAND_TYPE_READ_MULTICAST_GROUPS = 3;
  
  public static final short COMMAND_TYPE_ADD_DEVICES_MULTICAST_GROUPS = 4;
  
  public static final short COMMAND_TYPE_REMOVE_DEVICES_MULTICAST_GROUPS = 5;
  
  public static final short COMMAND_TYPE_SET_MULTICAST_GROUPS = 6;
  
  public static final short COMMAND_TYPE_BLACKLIST_CHANNEL = 7;
  
  public static final short COMMAND_TYPE_REMOVE_CHANNEL_BLACKLIST = 8;
  
  public static final short COMMAND_TYPE_READ_BLACKLISTED_CHANNELS = 9;
  
  public static final short COMMAND_TYPE_SET_BLACKLISTED_CHANNELS = 10;
  
  public static final short COMMAND_TYPE_FPGA_SET_REGISTER = 11;
  
  public static final short COMMAND_TYPE_FPGA_READ_REGISTER = 12;
  
  public static final short COMMAND_TYPE_SET_SIB = 13;
  
  public static final short COMMAND_TYPE_READ_SIB = 14;
  
  public static final short COMMAND_TYPE_SET_SDR_PARAMETER = 15;
  
  public static final short COMMAND_TYPE_PARAMETER = 16;
  
  public static final short COMMAND_TYPE_AT_COMMAND = 17;
  
  public static final short COMMAND_TYPE_SET_SCHEDULE_TRANSFER = 18;
  
  public static final short COMMAND_TYPE_GET_SCHEDULE_TRANSFER = 19;
  
  public static final short COMMAND_TYPE_RUN_MODE = 20;
  
  public static final short COMMAND_TYPE_STRING_PARAMETER = 21;
  
  public static final short COMMAND_TYPE_SAVE_CURRENT_STATE_AS_DEFAULT = 23;
  
  public static final short COMMAND_TYPE_SET_FREQ_SCAN = 24;
  
  public static final short COMMAND_TYPE_BS_FIRMWARE_UPGRADE = 25;
  
  public static final short COMMAND_TYPE_PING_REQUEST = 32;
  
  public static final short COMMAND_TYPE_ED_INPUT_QUEUE_SIZE = 33;
  
  public static final short COMMAND_TYPE_MCS_SETTINGS = 34;
  
  public static final short COMMAND_TYPE_MCS_VALUES = 35;
  
  public static final short COMMAND_TYPE_ED_RA_SLOTS = 36;
  
  // Can either be logs or anyfile in the future.
  public static final short COMMAND_TYPE_GET_LOGS = 46;
  //Not used in ENBC
  public static final short COMMAND_TYPE_GET_DL_EARFCN = 57;
  public static final short COMMAND_TYPE_GET_UL_EARFCN = 58;

  public static final short COMMAND_TYPE_PUT_FILE_EX = 47;
  
  public static final short COMMAND_TYPE_UNDEFINED = -1;
  
  public static final int BYTES_FOR_ELEMENTS_COUNT = 2;
  
  public static final String JSON_KEY_CHANNEL = "channel";
  
  public PSCommand(int messageId, int commandId, byte[] messageData, short commandType, byte typeCode) {
    super(messageId, commandId, messageData, typeCode);
    this.commandType = commandType;
    this.response = null;
    this.delegate = null;
  }
  
  public PSCommand(int messageId, int commandId, byte[] messageData, short commandType) {
    this(messageId, commandId, messageData, commandType, (byte)4);
    this.delegate = null;
  }
  
  public PSCommand(byte[] messageData, short commandType, PSCommandDelegate delegate) {
    super(getNextCommandId(), messageData, (byte)4);
    this.commandType = commandType;
    this.response = null;
    this.delegate = delegate;
  }
  
  private static int getNextCommandId() {
    int nextMid = -1;
    synchronized (commandCounterLock) {
      nextMid = commandIdCounter;
      commandIdCounter++;
    } 
    return nextMid;
  }
  
  protected byte[] getMessageDataBytes() throws MessageFormatException {
    short commandId = (short)(int)(this.applicationId & 0xFFFFL);
    ByteBuffer buffer = ByteBuffer.allocate(4 + this.data.length);
    buffer.order(ByteOrder.LITTLE_ENDIAN);
    buffer.putShort(commandId);
    buffer.putShort(this.commandType);
    buffer.put(this.data);
    return buffer.array();
  }
  
  public void cleanData() {
    this.data = new byte[0];
  }
  
  public synchronized PSCommandResponse getResponse() {
    return this.response;
  }
  
  public synchronized void setResponse(PSCommandResponse response) {
    if (response.getCommandType() == getCommandType()) {
      this.response = response;
      if (this.delegate != null)
        this.delegate.commandandReceivedResponse(this); 
    } else {
      System.out.println("Missmatch in command types. This command is type " + getCommandType() + " and response is " + response.getCommandType());
    } 
  }
  
  public PSCommandDelegate getDelegate() {
    return this.delegate;
  }
  
  public void setDelegate(PSCommandDelegate delegate) {
    this.delegate = delegate;
  }
  
  public abstract String getTypeString();
  
  public abstract String getContentString();
  
  public String getStatusStr() {
    if (this.response == null)
      return "No Response"; 
    if (this.response.isSuccess())
      return "Done"; 
    return "Failed";
  }
  
  public short getCommandType() {
    return this.commandType;
  }
  
  public String toString() {
    return getTypeString() + " " + getTypeString();
  }
}
