package ubiik.wp.psconnection;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.net.InetAddress;
import java.net.InterfaceAddress;
import java.net.NetworkInterface;
import java.net.SocketTimeoutException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import ubiik.utils.DatatypeConverter;
import ubiik.wp.model.messages.PSMessage;
import ubiik.wp.model.messages.PSMessageDiscovery;
import ubiik.wp.model.messages.PSMessageFactory;
import ubiik.wp.util.ApplicationLogger;
import ubiik.wp.view.MainViewController;

public class UUBaseFinder extends PSConnection {
  private static final byte[] DATA_MASK_FOR_DISCOVERY = new byte[] { 
      -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 
      -1, -1, -1, -1, -1, -1 };
  
  private HashMap<String, InetAddress> baseStationIp = null;
  
  private boolean processInput(DatagramPacket receivePacket) throws PSConnectionException {
    byte[] data = receivePacket.getData();
    int dataLength = receivePacket.getLength();
    byte[] nextMessage = new byte[dataLength - 2 - 4];
    ByteBuffer inBuffer = ByteBuffer.wrap(data).order(ByteOrder.LITTLE_ENDIAN);
    inBuffer.position(0);
    int messageLength = inBuffer.getInt();
    if (messageLength != dataLength - 2 - 4)
      throw new PSConnectionException(); 
    inBuffer.get(nextMessage, 0, nextMessage.length);
    inBuffer.position(dataLength - 2);
    short messageCrc = inBuffer.getShort();
    short calculatedCRC = crc16(nextMessage);
    if (messageCrc != calculatedCRC) {
      ApplicationLogger.debug("UDP Discovery: invalid CRC");
      throw new PSConnectionException();
    } 
    PSMessage message = PSMessageFactory.createMessageFromBytes(nextMessage);
    if (message != null && message instanceof PSMessageDiscovery) {
      byte[] bsId = ((PSMessageDiscovery)message).getDeviceId();
      this.baseStationIp.put(DatatypeConverter.printHexBinary(bsId), receivePacket.getAddress());
    } else {
      throw new PSConnectionException();
    } 
    return (inBuffer.capacity() > 0);
  }
  
  private byte[] buildDiscoveryRequest() {
    PSMessageDiscovery discovery = new PSMessageDiscovery(0, DATA_MASK_FOR_DISCOVERY);
    return buildPkt(discovery.toBytes());
  }
  
  public void findBaseStations(int port, int timeoutMilliseconds) {
    this.baseStationIp = new HashMap<>();
    try {
      DatagramSocket c = new DatagramSocket();
      c.setBroadcast(true);
      byte[] sendData = buildDiscoveryRequest();
      try {
        DatagramPacket sendPacket = new DatagramPacket(sendData, sendData.length, InetAddress.getByName("255.255.255.255"), port);
        c.send(sendPacket);
      } catch (Exception exception) {}
      Enumeration<NetworkInterface> interfaces = NetworkInterface.getNetworkInterfaces();
      while (interfaces.hasMoreElements()) {
        NetworkInterface networkInterface = interfaces.nextElement();
        if (networkInterface.isLoopback() || !networkInterface.isUp())
          continue; 
        for (InterfaceAddress interfaceAddress : networkInterface.getInterfaceAddresses()) {
          InetAddress broadcast = interfaceAddress.getBroadcast();
          if (broadcast == null)
            continue; 
          try {
            DatagramPacket sendPacket = new DatagramPacket(sendData, sendData.length, broadcast, port);
            c.send(sendPacket);
          } catch (Exception exception) {}
        } 
      } 
      ApplicationLogger.debug(getClass().getName() + ">>> Done looping over all network interfaces. Waiting for a reply.");
      int timeRemaining = timeoutMilliseconds;
      long endTimeMillis = System.currentTimeMillis() + timeoutMilliseconds;
      byte[] recvBuf = new byte[1024];
      while (timeRemaining > 0) {
        DatagramPacket receivePacket = new DatagramPacket(recvBuf, recvBuf.length);
        c.setSoTimeout(timeRemaining);
        try {
          c.receive(receivePacket);
          boolean validInput = processInput(receivePacket);
          String message = (new String(receivePacket.getData())).trim();
          ApplicationLogger.debug("Received " + message + "from: " + receivePacket.getAddress());
          if (validInput)
            ApplicationLogger.debug("FOUND BS with IP address: " + receivePacket.getAddress()); 
          timeRemaining = (int)(endTimeMillis - System.currentTimeMillis());
        } catch (SocketTimeoutException e) {
          timeRemaining = 0;
        } catch (Exception e) {
          ApplicationLogger.debug("Discovery received Invalid packet");
        } 
      } 
      c.close();
    } catch (IOException ex) {
      Logger.getLogger(MainViewController.class.getName()).log(Level.SEVERE, (String)null, ex);
    } 
  }
  
  public HashMap<String, InetAddress> getResults() {
    return this.baseStationIp;
  }
}
