package ubiik.wp.view;

import javafx.application.Platform;
import javafx.css.PseudoClass;
import javafx.scene.control.Alert;
import javafx.scene.control.Label;
import javafx.scene.control.TextArea;
import javafx.scene.control.TextField;
import ubiik.wp.controller.DevicesController;
import ubiik.wp.controller.MessageController;
import ubiik.wp.model.messages.PSMessage;
import ubiik.wp.model.messages.pscommands.PSCommand;

public abstract class BaseViewController extends RefreshableViewController {
  protected DevicesController devicesController;
  
  protected MessageController messageController;
  
  protected CommandStatusReceiver commandStatusReceiver;
  
  public BaseViewController(DevicesController devicesController, MessageController messageController, CommandStatusReceiver commandStatusReceiver) {
    this.devicesController = devicesController;
    this.messageController = messageController;
    this.commandStatusReceiver = commandStatusReceiver;
  }
  
  protected void setTextFieldHasError(TextField textField, boolean hasError) {
    PseudoClass errorClass = PseudoClass.getPseudoClass("error");
    textField.pseudoClassStateChanged(errorClass, hasError);
  }
  
  protected void setTextFieldIsSuccessful(TextField textField, boolean successful) {
    PseudoClass errorClass = PseudoClass.getPseudoClass("success");
    textField.pseudoClassStateChanged(errorClass, successful);
  }
  
  protected void sendCommandAndUpdateStatusBar(PSCommand command, boolean setBar) {
    this.messageController.sendMessage((PSMessage)command);
    if (setBar)
      this.commandStatusReceiver.setLastStatus("Sent command " + command.getTypeString()); 
    command.cleanData();
  }
  
  protected void sendCommandAndUpdateStatusBar(PSCommand command) {
    sendCommandAndUpdateStatusBar(command, false);
  }
  
  protected void setStatusBarCommandFailed(PSCommand command) {
    this.commandStatusReceiver.setLastStatus("Command " + command.getTypeString() + " failed");
  }
  
  protected void setStatusBarCommandSuccess(PSCommand command) {
    this.commandStatusReceiver.setLastStatus("Command executed succesfully: " + command
        .getTypeString() + "; " + command.getContentString());
  }
  
  protected void displayErrorMessage(String title, String error) {
    displayMessage(title, error, Alert.AlertType.ERROR);
  }
  
  protected void displayWarningMessage(String title, String message) {
    displayMessage(title, message, Alert.AlertType.WARNING);
  }
  
  protected void displayInfoMessage(String title, String message) {
    displayMessage(title, message, Alert.AlertType.INFORMATION);
  }
  
  protected void displayMessage(final String title, final String message, final Alert.AlertType alertType) {
    Platform.runLater(new Runnable() {
          public void run() {
            Alert alert = new Alert(alertType);
            alert.setTitle(title);
            alert.setContentText(message);
            alert.showAndWait();
          }
        });
  }
  
  protected Integer validateAndGetIntegerTextFieldMinMax(TextField textField, int min, int max) {
    boolean error = false;
    int value = -1;
    try {
      value = Integer.parseInt(textField.getText());
      if (value < min || value > max)
        error = true; 
    } catch (Exception e) {
      error = true;
    } 
    setTextFieldHasError(textField, error);
    if (!error)
      return Integer.valueOf(value); 
    return null;
  }
  
  protected Integer validateAndGetIntegerTextFieldSetValues(TextField textField, Integer[] values) {
    boolean error = false;
    int value = -1;
    boolean match = false;
    try {
      value = Integer.parseInt(textField.getText());
      for (int i = 0; i < values.length && !match; i++)
        match = (value == values[i].intValue()); 
    } catch (Exception e) {
      error = true;
    } 
    setTextFieldHasError(textField, (error || !match));
    if (!error && match)
      return Integer.valueOf(value); 
    return null;
  }
  
  protected void cleanInput(TextField input) {
    cleanInput(input, "");
  }
  
  protected void cleanInput(final TextField input, final String text) {
    Platform.runLater(new Runnable() {
          public void run() {
            input.setText(text);
          }
        });
  }
  
  protected void setLabelText(final Label label, final String text) {
    Platform.runLater(new Runnable() {
          public void run() {
            label.setText(text);
          }
        });
  }
  
  protected void appendToTextArea(final TextArea textArea, final String text) {
    Platform.runLater(new Runnable() {
          public void run() {
            textArea.appendText(text);
          }
        });
  }
}
