package ubiik.wp.view;

import java.net.InetAddress;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.LinkedBlockingQueue;
import javafx.application.Platform;
import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.ButtonType;
import javafx.scene.control.ChoiceDialog;
import javafx.scene.control.Label;
import javafx.scene.control.TextField;
import javafx.scene.layout.BorderPane;
import ubiik.utils.DatatypeConverter;
import ubiik.wp.controller.BaseStationNameManager;
import ubiik.wp.controller.ConnectionObserver;
import ubiik.wp.controller.DevicesController;
import ubiik.wp.controller.MessageController;
import ubiik.wp.controller.UserManager;
import ubiik.wp.psconnection.PSConnectionErrorCode;
import ubiik.wp.psconnection.PSConnectionManager;
import ubiik.wp.psconnection.PSConnectionManagerDelegate;
import ubiik.wp.psconnection.SocketPSConnectionManager;
import ubiik.wp.psconnection.UUBaseFinder;
import ubiik.wp.psconnection.auth.PSAuth;

public abstract class ConnectionManagerViewController extends BaseViewController implements ConnectionObserver {
  private static final String BASE_ID_IP_SEPARATOR = " | ";
  
  private static final int FIND_WAIT_TIME = 2500;
  
  @FXML
  protected BorderPane rootPane;
  
  @FXML
  protected Label labelConnectionStatus;
  
  @FXML
  protected Button connectionButton;
  
  @FXML
  protected TextField labelBaseStationId;
  
  @FXML
  protected TextField inputHost;
  
  @FXML
  protected TextField inputPort;
  
  @FXML
  protected Label labelStatus;
  
  protected PSAuth user;
  
  protected boolean displayInputPort;
  
  public ConnectionManagerViewController(MessageController mc, DevicesController dc) {
    super(dc, mc, null);
    this.user = UserManager.getInstance().getPsAuth();
    this.displayInputPort = false;
  }
  
  public void initialize() {
    this.inputHost.setText("127.0.0.1");
    this.inputPort.setText(Integer.toString(7878));
    this.inputPort.managedProperty().bind(this.inputPort.visibleProperty());
    this.inputPort.setVisible(this.displayInputPort);
    this.messageController.addConnectionObserver(this);
  }
  
  public PSConnectionManager getNewPSConnectionManager(String host, int port) {
    return (PSConnectionManager)new SocketPSConnectionManager(host, port, (PSConnectionManagerDelegate)this.messageController, new LinkedBlockingQueue());
  }
  
  public void connectionFailed(PSConnectionErrorCode errorCode) {
    authenticationFinished(false);
  }
  
  private void restartStorages() {
    this.messageController.restartMessageStorage();
    this.devicesController.restart(true, true);
  }
  
  public void updateConnectionStatus(final String status) {
    Platform.runLater(new Runnable() {
          public void run() {
            ConnectionManagerViewController.this.labelConnectionStatus.setText(status);
          }
        });
  }
  
  public void updateConnectionStatus(final boolean connected) {
    Platform.runLater(new Runnable() {
          public void run() {
            if (connected) {
              ConnectionManagerViewController.this.labelConnectionStatus.setText("Connected");
              ConnectionManagerViewController.this.connectionButton.setText("Disconnect");
              byte[] bsid = ConnectionManagerViewController.this.devicesController.getBaseStationId();
              String bsidStr = DatatypeConverter.printHexBinary(bsid);
              String bsName = BaseStationNameManager.getInstance().getNameForUUID(bsidStr);
              if (bsName != null)
                bsidStr = bsName; 
              ConnectionManagerViewController.this.labelBaseStationId.setText("Base Station " + bsidStr);
            } else {
              ConnectionManagerViewController.this.labelConnectionStatus.setText("Disconnected");
              ConnectionManagerViewController.this.connectionButton.setText("Connect");
              ConnectionManagerViewController.this.labelBaseStationId.setText("");
            } 
            ConnectionManagerViewController.this.setConnectionParamsEditable(!connected);
            ConnectionManagerViewController.this.connectionButton.setDisable(false);
          }
        });
  }
  
  private void setConnectionParamsEditable(boolean editable) {
    this.inputHost.setEditable(editable);
    this.inputPort.setEditable(editable);
  }
  
  private void setConnectionParamsVisible(boolean visible) {
    this.inputHost.managedProperty().bind(this.inputHost.visibleProperty());
    this.inputHost.setVisible(visible);
    this.inputPort.managedProperty().bind(this.inputPort.visibleProperty());
    this.inputPort.setVisible(visible & this.displayInputPort);
  }
  
  public void authenticationFinished(boolean autheticated) {
    disableUI(false);
    disableMenuAndActions(!autheticated);
    updateConnectionStatus(autheticated);
    if (!autheticated) {
      displayErrorMessage("Connection Failure", "Connection failure. Please check network environment.");
      setLastStatusLabel("Connection Failure");
    } else {
      setLastStatusLabel("Connected to base station");
      restartStorages();
    } 
    updateConnectionStatus(autheticated);
  }
  
  public void setLastStatusLabel(final String lastStatus) {
    Platform.runLater(new Runnable() {
          public void run() {
            ConnectionManagerViewController.this.labelStatus.setText(lastStatus);
          }
        });
  }
  
  @FXML
  public void discoverBSPressed() {
    final UUBaseFinder baseFinder = new UUBaseFinder();
    boolean portValid = validatePort();
    if (portValid) {
      final int port = Integer.parseInt(this.inputPort.getText());
      final String connectioStatus = this.labelConnectionStatus.getText();
      this.labelConnectionStatus.setText("Searching");
      setLastStatusLabel("Searching for base stations in local network");
      disableUI(true);
      Thread t = new Thread(new Runnable() {
            public void run() {
              baseFinder.findBaseStations(port, 2500);
              Platform.runLater(new Runnable() {
                    public void run() {
                      ConnectionManagerViewController.this.finishedBSDiscovery(baseFinder, connectioStatus);
                    }
                  });
            }
          });
      t.start();
    } 
  }
  
  public void disableUI(boolean disable) {
    this.rootPane.setDisable(disable);
  }
  
  public void disconnect() {
    if (this.messageController.isConnected().get()) {
      this.messageController.disconnect();
      if (this.devicesController.isInactiveDevicesMonitoringEnabled())
        this.devicesController.stopTrackingInactiveDevices(); 
      updateConnectionStatus(false);
      disableMenuAndActions(true);
      setLastStatusLabel("Disconnected from Base Station");
    } 
  }
  
  public void connect() {
    Platform.runLater(new Runnable() {
          public void run() {
            final String host = ConnectionManagerViewController.this.inputHost.getText();
            final int port = Integer.parseInt(ConnectionManagerViewController.this.inputPort.getText());
            ConnectionManagerViewController.this.labelConnectionStatus.setText("Connecting");
            ConnectionManagerViewController.this.setLastStatusLabel("Connecting to " + host);
            ConnectionManagerViewController.this.connectionButton.setDisable(true);
            ConnectionManagerViewController.this.disableUI(true);
            Thread t = new Thread(new Runnable() {
                  public void run() {
                    ConnectionManagerViewController.this.messageController.connect(host, port, ConnectionManagerViewController.this.getNewPSConnectionManager(host, port));
                  }
                });
            t.start();
          }
        });
  }
  
  @FXML
  public void connectionButtonPressed() {
    if (this.messageController.isConnected().get()) {
      disconnect();
    } else if (validateHostAndPort()) {
      connect();
    } 
  }
  
  private boolean validatePort() {
    boolean portValid = true;
    try {
      int port = Integer.parseInt(this.inputPort.getText());
      portValid = (port > 0);
    } catch (Exception e) {
      portValid = false;
    } 
    setTextFieldHasError(this.inputPort, !portValid);
    return portValid;
  }
  
  private boolean validateHostAndPort() {
    boolean hostValid = true;
    boolean portValid = validatePort();
    hostValid = (this.inputHost.getText().length() > 0);
    setTextFieldHasError(this.inputHost, !hostValid);
    return (hostValid && portValid);
  }
  
  public void finishedBSDiscovery(UUBaseFinder baseFinder, String prevConnectioStatus) {
    updateConnectionStatus(prevConnectioStatus);
    HashMap<String, InetAddress> baseIps = baseFinder.getResults();
    int totalBs = baseIps.size();
    if (totalBs > 0) {
      List<String> choices = new ArrayList<>();
      for (String baseId : baseIps.keySet())
        choices.add(baseId + " | " + baseId); 
      ChoiceDialog<String> dialog = new ChoiceDialog<>(choices.get(0), choices);
      Button okButton = (Button)dialog.getDialogPane().lookupButton(ButtonType.OK);
      okButton.setText("Connect");
      dialog.setTitle("Choose a Base Station");
      String pl = (totalBs != 1) ? "s" : "";
      dialog.setHeaderText("Found " + totalBs + " Base Station" + pl + ". Please select the one you would like to connect to");
      setLastStatusLabel("Finished searching for base stations.Found " + totalBs);
      Optional<String> result = dialog.showAndWait();
      if (result.isPresent()) {
        String strResult = result.get();
        String baseId = strResult.substring(0, strResult.indexOf(" | ".charAt(0)));
        this.inputHost.setText(((InetAddress)baseIps.get(baseId)).getCanonicalHostName());
        connectionButtonPressed();
      } 
    } else {
      displayErrorMessage("No Base Station Found", "No Base Station found in your local network");
      setLastStatusLabel("Finished searching for base stations. No Base Station Found.");
    } 
    disableUI(false);
  }
  
  public abstract void disableMenuAndActions(boolean paramBoolean);
}
