package ubiik.wp.view;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.List;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.css.PseudoClass;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.Parent;
import javafx.scene.control.TextField;
import javafx.scene.layout.VBox;
import ubiik.utils.DatatypeConverter;
import ubiik.wp.controller.DevicesController;
import ubiik.wp.controller.MessageController;
import ubiik.wp.model.fpga.FPGARegister;
import ubiik.wp.model.fpga.FPGARegisterField;
import ubiik.wp.model.fpga.FPGARegisterObserver;
import ubiik.wp.model.messages.pscommands.PSCommand;
import ubiik.wp.model.messages.pscommands.PSCommandDelegate;
import ubiik.wp.model.messages.pscommands.PSCommandFPGAReadRegisters;
import ubiik.wp.model.messages.pscommands.PSCommandFPGASetRegister;
import ubiik.wp.model.messages.pscommands.responses.PSCommandResponseFPGAReadRegister;
import ubiik.wp.util.ApplicationLogger;

public class FPGARegisterViewController extends BaseViewController implements FPGARegisterObserver, PSCommandDelegate {
  public static final int BYTES_PER_REGISTER = 4;

  @FXML
  private TextField regValueTextField;

  @FXML
  private VBox fieldsVBOX;

  private FPGARegister register;

  private List<FPGARegisterFieldController> fieldControllers;

  public FPGARegisterViewController(FPGARegister register, MessageController messageController, DevicesController devicesController, CommandStatusReceiver commandStatusReceiver) {
    super(devicesController, messageController, commandStatusReceiver);
    this.register = register;
    this.fieldControllers = new ArrayList<>();
    this.register.addObserver(this);
  }

  public void initialize() {
    setRegisterInputText();
    for (FPGARegisterField field : this.register.getFields()) {
      try {
        FPGARegisterFieldController fieldVC = new FPGARegisterFieldController(field);
        FXMLLoader fpgaRegisterFieldLoader = new FXMLLoader(getClass().getResource("/main/resources/ubiik/wp/admin/FPGARegisterFieldView.fxml"));
        fpgaRegisterFieldLoader.setController(fieldVC);
        Parent registerPane = fpgaRegisterFieldLoader.<Parent>load();
        this.fieldsVBOX.getChildren().add(registerPane);
        this.fieldControllers.add(fieldVC);
      } catch (IOException e) {
        e.printStackTrace();
      }
    }
    this.regValueTextField.focusedProperty()
      .addListener(new ChangeListener<Boolean>() {
          public void changed(ObservableValue<? extends Boolean> arg0, Boolean oldPropertyValue, Boolean newPropertyValue) {
            if (!newPropertyValue.booleanValue())
              FPGARegisterViewController.this.validateInputAndSet();
          }
        });
  }

  private void validateInputAndSet() {
    boolean valid = validateInput(true);
    if (valid)
      setRegisterInputText();
  }

  private void setRegisterInputText() {
    ByteBuffer buffer = ByteBuffer.allocate(4);
    int int32val = this.register.getAs32Bit();
    buffer.order(ByteOrder.BIG_ENDIAN);
    buffer.putInt(int32val);
    String str = DatatypeConverter.printHexBinary(buffer.array());
    this.regValueTextField.setText(str);
    for (FPGARegisterFieldController fpgaRegisterFieldController : this.fieldControllers)
      fpgaRegisterFieldController.setFrom32Bit(int32val);
  }

  public boolean validateInput(boolean save) {
    boolean valid = false;
    int intValue = 0;
    String str = this.regValueTextField.getText();
    if (str.length() == 8) {
      try {
        intValue = Integer.parseUnsignedInt(str, 16);
        valid = true;
      } catch (Exception e) {
        valid = false;
      }
    } else {
      ApplicationLogger.debug("FPGA validateInput: str length invalid");
    }
    if (valid && save)
      for (FPGARegisterFieldController fpgaRegisterFieldController : this.fieldControllers)
        fpgaRegisterFieldController.setFrom32Bit(intValue);
    setFieldWithError(!valid);
    return valid;
  }

  private void setFieldWithError(boolean error) {
    PseudoClass errorClass = PseudoClass.getPseudoClass("error");
    this.regValueTextField.pseudoClassStateChanged(errorClass, error);
  }

  @FXML
  public void setButtonPressed() {
    validateInputAndSet();
  }

  @FXML
  public void writeButtonPressed() {
    boolean valid = true;
    for (FPGARegisterFieldController fieldVC : this.fieldControllers)
      valid = (valid && fieldVC.validateInput(true));
    if (valid) {
      PSCommandFPGASetRegister cmd = new PSCommandFPGASetRegister(this.register, this);
      cmd.setDelegate(this);
      sendCommandAndUpdateStatusBar((PSCommand)cmd);
    }
  }

  private void setRegisterFromResponse(PSCommandResponseFPGAReadRegister response) {
    if (response.getRegisterIdx() == -1) {
      this.devicesController.setAllFPGARegisters(response.getRegisterContent());
    } else if (response.getRegisterIdx() == (byte)(0xFF & this.register.getIndex())) {
      if ((response.getRegisterContent()).length == 4) {
        ByteBuffer buff = ByteBuffer.wrap(response.getRegisterContent()).order(ByteOrder.LITTLE_ENDIAN);
        this.register.setFrom32Bit(buff.getInt());
      } else {
        System.err.println("Discard command response. Received " + (response
            .getRegisterContent()).length + " bytes instead of 4");
      }
    } else {
      System.err.println("Discard command response. Register indexes don't match");
    }
  }

  @FXML
  public void readButtonPressed() {
    sendCommandAndUpdateStatusBar((PSCommand)new PSCommandFPGAReadRegisters(this.register, this));
  }

  public void registerChanged() {
    setRegisterInputText();
    setFieldWithError(false);
  }

  public void commandandReceivedResponse(PSCommand command) {
    if (command.getResponse().isSuccess()) {
      setStatusBarCommandSuccess(command);
      if ((command instanceof PSCommandFPGAReadRegisters && command
        .getResponse() instanceof PSCommandResponseFPGAReadRegister) || (command instanceof PSCommandFPGASetRegister && command

        .getResponse() instanceof ubiik.wp.model.messages.pscommands.responses.PSCommandResponseFPGASetRegister)) {
        PSCommandResponseFPGAReadRegister resp = (PSCommandResponseFPGAReadRegister)command.getResponse();
        setRegisterFromResponse(resp);
      }
    } else {
      displayErrorMessage("Error", "The command failed");
      setStatusBarCommandFailed(command);
    }
  }

  public int getRegisterContent() {
    return this.register.getAs32Bit();
  }

  public void doReloadView() {}
}
