package ubiik.wp.view;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import javafx.application.Platform;
import javafx.event.ActionEvent;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.Node;
import javafx.scene.control.Alert;
import javafx.scene.control.Button;
import javafx.scene.control.ButtonType;
import javafx.scene.control.ChoiceBox;
import javafx.scene.control.Label;
import javafx.scene.control.TextArea;
import javafx.scene.layout.Pane;
import ubiik.wp.controller.DevicesController;
import ubiik.wp.controller.MessageController;
import ubiik.wp.controller.PSMessageStorage;
import ubiik.wp.controller.TestModesManager;
import ubiik.wp.controller.UserManager;
import ubiik.wp.model.bsparameters.BSParameter;
import ubiik.wp.model.messages.PSLogMessage;
import ubiik.wp.model.messages.pscommands.PSCommand;
import ubiik.wp.model.messages.pscommands.PSCommandDelegate;
import ubiik.wp.model.messages.pscommands.PSCommandParameter;
import ubiik.wp.model.messages.pscommands.PSCommandRunMode;
import ubiik.wp.model.messages.pscommands.responses.PSCommandResponse;
import ubiik.wp.model.testmode.TestModesListener;
import ubiik.wp.model.testmode.UUTestMode;

public class TestModeViewController extends BaseViewController implements PSCommandDelegate, TestModesListener {
  @FXML
  protected TextArea textArea;

  @FXML
  protected Pane testModeBtnsContainer;

  private Date lastTimestamp;

  private List<UUTestMode> testModes;

  private ChoiceBox<UUTestMode> choiceboxTestMode;

  private ArrayList<BSParameterViewController> bsParameterVCs;

  @FXML
  private Pane bsParametersContainer;

  public TestModeViewController(DevicesController devicesController, MessageController messageController, CommandStatusReceiver commandStatusReceiver) {
    super(devicesController, messageController, commandStatusReceiver);
    this.lastTimestamp = null;
    this.bsParameterVCs = new ArrayList<>();
  }

  public void initialize() {
    TestModesManager testModesManager = TestModesManager.getInstance();
    testModesManager.addListener(this);
    this.testModes = testModesManager.getTestModes();
    if (this.testModes.isEmpty()) {
      testModesManager.setMessageController(this.messageController);
      testModesManager.refreshTestModes();
    } else {
      setUpTestModesUI();
    }
    if (UserManager.getInstance().getPsAuth()
      .isCommandAvailable(PSCommandParameter.class))
      loadBSParameters();
  }

  private void loadBSParameters() {
    ArrayList<BSParameter> parameters = this.devicesController.getBSParameters();
    for (int i = 0; i < parameters.size(); i++) {
      if (((BSParameter)parameters.get(i)).getCategory().compareTo("UT") == 0) {
        BSParameterViewController vc = new BSParameterViewController(this.devicesController, this.messageController, this.commandStatusReceiver, i);
        this.bsParameterVCs.add(vc);
        FXMLLoader loader = new FXMLLoader(getClass().getResource("/main/java/ubiik/wp/view/BSParameterView.fxml"));
        loader.setController(vc);
        try {
          this.bsParametersContainer.getChildren().add(loader.<Node>load());
        } catch (IOException e) {
          e.printStackTrace();
        }
      }
    }
  }

  public void btnStartTestModePressed() {
    UUTestMode testMode = this.choiceboxTestMode.getValue();
    if (testMode != null) {
      sendCommandAndUpdateStatusBar((PSCommand)new PSCommandRunMode(this, (byte)1, testMode.getRunMode()));
    } else {
      displayErrorMessage("Select Test Mode", "Please select a Test Mode from the list.");
    }
  }

  @FXML
  public void btnStartPressed() {
    Alert alert = new Alert(Alert.AlertType.CONFIRMATION);
    alert.setTitle("Restart Protocol Stack");
    alert.setHeaderText("This will cause the Protocol Stack to restart and this application will be disconnected.");
    alert.setContentText("Continue?");
    Optional<ButtonType> result = alert.showAndWait();
    if (result.isPresent() && result.get() == ButtonType.OK)
      sendCommandAndUpdateStatusBar((PSCommand)new PSCommandRunMode(this, (byte)1, 1));
  }

  @FXML
  public void btnStopPressed() {
    sendCommandAndUpdateStatusBar((PSCommand)new PSCommandRunMode(this, (byte)1, 0));
  }

  public void commandandReceivedResponse(PSCommand command) {
    PSCommandResponse resp = command.getResponse();
    if (resp.isSuccess()) {
      setStatusBarCommandSuccess(command);
      if (command instanceof PSCommandRunMode && resp instanceof ubiik.wp.model.messages.pscommands.responses.PSCommandResponseRunMode) {
        String runModeStr = ((PSCommandRunMode)command).getRunModeStr();
        if (runModeStr.compareTo("") != 0)
          appendToTextArea(this.textArea, "> " + runModeStr + "\n");
      }
    } else {
      this.textArea.appendText("\nCommand Failed");
      appendToTextArea(this.textArea, "\nCommand Failed");
      setStatusBarCommandFailed(command);
    }
  }

  private void setUpTestModesUI() {
    Platform.runLater(new Runnable() {
          public void run() {
            TestModeViewController.this.testModeBtnsContainer.getChildren().clear();
            if (TestModeViewController.this.testModes.isEmpty()) {
              Label label = new Label();
              label.setText("No Test Modes available");
              TestModeViewController.this.testModeBtnsContainer.getChildren().add(label);
              TestModeViewController.this.choiceboxTestMode = null;
            } else {
              TestModeViewController.this.choiceboxTestMode = new ChoiceBox<>();
              TestModeViewController.this.choiceboxTestMode.getItems().addAll(TestModeViewController.this.testModes);
              TestModeViewController.this.choiceboxTestMode.getSelectionModel().select(0);
              Button testModeBtn = new Button();
              testModeBtn.setOnAction(e -> TestModeViewController.this.btnStartTestModePressed());
              testModeBtn.setText("Start Test");
              TestModeViewController.this.testModeBtnsContainer.getChildren().addAll(new Node[] { TestModeViewController.this.choiceboxTestMode , (Node)testModeBtn });
            }
          }
        });
  }

  public void updateTestModes() {
    List<UUTestMode> nTestModes = TestModesManager.getInstance().getTestModes();
    this.testModes = nTestModes;
    setUpTestModesUI();
  }

  public void updateTestModesFailed() {
    displayErrorMessage("Error", "Failed to get the list of available Test Modes");
  }

  public void doReloadView() {
    PSMessageStorage ms = PSMessageStorage.getInstance();
    List<PSLogMessage> logs = ms.getLogs();
    int logsSize = logs.size();
    if (logsSize > 0) {
      List<PSLogMessage> subList = null;
      if (this.lastTimestamp != null) {
        subList = (List<PSLogMessage>)logs.stream().filter(elem -> (elem.getTimestamp().compareTo(this.lastTimestamp) > 0)).collect(Collectors.toList());
      } else {
        subList = logs;
      }
      for (PSLogMessage log : subList)
        appendToTextArea(this.textArea, log.getMessage() + "\n");
      if (subList.size() > 0)
        this.lastTimestamp = ((PSLogMessage)subList.get(subList.size() - 1)).getTimestamp();
    }
    for (BSParameterViewController bsParameterViewController : this.bsParameterVCs)
      bsParameterViewController.updateInput();
  }

  public void btnClearTerminalPressed() {
    this.textArea.clear();
  }
}
